"""
Link to Itential Docs: https://apidocs.itential.com/2020.2/api/app-admin_essentials/

Implemented  Doc String  Tests
    [x]         [/]       [ ]   getServicesHealth
"""
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from itential.core import Itential
    import requests


class AppAdminEssentials:
    """https://apidocs.itential.com/2020.2/api/app-admin_essentials/"""

    @staticmethod
    def get_services_health(client: "Itential") -> "requests.Response":
        """
        Gets the health of all services in an IAP environment.
        https://apidocs.itential.com/2020.2/api/app-admin_essentials/getServicesHealth/
        :param client: Itential state object
        :return:
        """
        return client.call(method="GET", url=f"{client.url}/admin/services/health")



# from itential.modules.version2021._app_form_builder import AppFormBuilder2021
#
#
# class AppFormBuilder2022(AppFormBuilder2021):
#     def cool_build_form_meth(*args, **kwargs):
#         print('build form meth 2022')



import logging
from typing import Dict, Any

import requests

logging.basicConfig(level=logging.WARNING)
log = logging.getLogger(__name__)


class Itential:
    def __init__(self, **kwargs) -> None:  # type: ignore
        """
        :param username: Know defaults include "admin@pronghorn" and "admin@itential"
        :param password: For authenticating with the server. Never stored or printed.
        :param version: Not in use yet, might help with using the correct methods for authentication in the future.
        :param url: Used to change the server. Default is the local machine.
        :param session: Pass in a previously saved requests.Session object in order to bypass the authentication call.
        """
        self.username: str = kwargs.get("username", "admin@pronghorn")
        self.password: str = kwargs.get("password", "admin")
        self.version: str = kwargs.get("version", "latest")
        self.auth_body: Dict[str, Dict[str, str]] = {"user": {"username": self.username, "password": self.password}}

        self._url: str = "http://localhost:3000"
        # self._url - Probably a better way to do this.
        # self.__setattr__('url', kwargs.get('url', "http://localhost:3000"))

        self.session: requests.Session = kwargs.get('session', requests.Session())

    @property
    def url(self) -> str:
        return self._url

    @url.setter
    def url(self, value: str) -> None:
        """Cleans the input URL of any extra whitespace and trailing slashes."""
        if value:
            trimmed_value = value.strip()
            if trimmed_value.endswith('/'):
                value = trimmed_value[::-1].replace('/', '', 1)[::-1]
            self._url = value

    def call(self, method: str, url: str, **kwargs: Any) -> requests.Response:
        """
        Wrapper for the Requests request method. Designed to have the Itential class authenticate with the server
        or verify it's connection before making the module call.
        """
        self.authenticate()

        headers = kwargs.get("headers", {"Content-Type": "application/json"})  # Required by some calls.
        verify = kwargs.get("verify", False)

        return self.session.request(method=method, url=url, headers=headers, verify=verify, **kwargs)

    def authenticate(self) -> None:
        """
        Authenticates with the specified server. Attempts to use self.session if available.
        Updates the self.session for use with subsequent calls.
        """
        # Make initial authentication check with old/previous session cookies.
        if not self.get_version_health_check().ok:
            log.debug('Initial Health Check failed. Authenticating with server.')
            self.verify_remote_auth_info()  # Verifies username and prompts for the password.
            self._authenticate_call()  # Authenticate with server

    def _authenticate_call(self) -> None:
        response = self.session.post(f"{self.url}/login/", json=self.auth_body, verify=False)
        if response.ok:
            self.password = "admin"  # Reset to default password for local access.
            log.debug(f"Authenticated with {self.url} server. [{response.status_code}]")
        else:
            log.error(
                f"{response.status_code} - Failed to authenticate with {self.url} server. "
                f"Reason: '{response.json()['message']}'"
            )

    def get_version_health_check(self) -> requests.Response:
        """
        Pulls a dictionary of all currently installed apps/adapters on the server, it's status, and versions.
        Also labeled as a health check in the IAP docs.
        Used as a quick test to verify the current session object before attempting to (re)authenticate.
        """
        response: requests.Response = self.session.get(url=f"{self.url}/health/applications")
        log.debug(f"Health check: {response.status_code}")
        return response

    def verify_remote_auth_info(self) -> None:
        """
        Holdover method from when this was a TUI app.
        Need to validate there's a valid username/password or not. If not, we'd use the default
        information for authentication, unless given something different.
        """
        if not self.username:
            raise ValueError("Must have a Username to authenticate with server.")
        log.debug(f"Username validated: '{self.username}'")

        if not self.password:
            assert isinstance(self.password, str), "A password is required."
        log.debug("Password validated.")



"""
Link to Itential Docs: https://apidocs.itential.com/2020.2/api/app-automation_catalog/

Implemented  Doc String  Tests
    [x]         [x]       [ ]   createAutomation
    [x]         [x]       [ ]   deleteAutomations
    [x]         [x]       [ ]   exportAutomation
    [x]         [x]       [ ]   getAutomationById
    [x]         [x]       [ ]   getAutomations
    [x]         [x]       [ ]   importAutomations
    [x]         [x]       [ ]   migrateAgendaJobs
    [x]         [x]       [ ]   runAutomation
    [x]         [x]       [ ]   updateAutomation
"""
from typing import TYPE_CHECKING, List, Dict, Any

if TYPE_CHECKING:
    from itential.core import Itential
    import requests


class AppAutomationCatalog:
    """https://apidocs.itential.com/2020.2/api/app-automation_catalog/"""

    @staticmethod
    def create_automation(client: "Itential", name: str, description: str) -> "requests.Response":
        """
        Creates an automation with optional description.
        An automation's root structure is based off the agenda.js library,
         with any custom data falling into the 'data' array
        https://apidocs.itential.com/2020.2/api/app-automation_catalog/createAutomation/
        :param client: Itential state object
        :param name: Unique name of the automation
        :param description: Short description of the automation
        :return: requests.Response: Automation document that was created by the request
        """
        body = {"name": name, "description": description}
        return client.call(method="POST", url=f"{client.url}/automation_catalog/automations", json=body)

    @staticmethod
    def delete_automations(client: "Itential", ids: List[str]) -> "requests.Response":
        """
        Deletes automations (based off an array of Ids)
        https://apidocs.itential.com/2020.2/api/app-automation_catalog/deleteAutomations/
        :param client: Itential state object
        :param ids: List of automation IDs
        :return: requests.Response: Object containing a property removeCount
        """
        body = {"ids": ids}
        return client.call(method="DELETE", url=f"{client.url}/automation_catalog/automations", json=body)

    @staticmethod
    def export_automations(client: "Itential", automation_id: str) -> "requests.Response":
        """
        Returns a single automation formatted for importing
        https://apidocs.itential.com/2020.2/api/app-automation_catalog/exportAutomation/
        :param client: Itential state object
        :param automation_id: Unique id of the automation
        :return: requests.Response: Automation document that was requested in exported format
        """
        return client.call(method="GET", url=f"{client.url}/automation_catalog/automations/{automation_id}/export")

    @staticmethod
    def get_automation_by_id(client: "Itential", automation_id: str) -> "requests.Response":
        """
        Gets an single automation by its id
        https://apidocs.itential.com/2020.2/api/app-automation_catalog/getAutomationById/
        :param client: Itential state object
        :param automation_id: Unique id of the automation
        :return: requests.Response: Automation document that was requested
        """
        return client.call(method="GET", url=f"{client.url}/automation_catalog/automations/{automation_id}")

    @staticmethod
    def get_automations(client: "Itential", query_parameters: Dict[str, Any]) -> "requests.Response":
        """
        Gets all known automations returned in alphabetical order
        https://apidocs.itential.com/2020.2/api/app-automation_catalog/getAutomations/
        :param client: Itential state object
        :param query_parameters: < Undocumented >
        :return: requests.Response: List of all automation documents
        """
        params = {"queryParameters": query_parameters}
        return client.call(method="GET", url=f"{client.url}/automation_catalog/automations", params=params)

    @staticmethod
    def import_automations(
        client: "Itential", automation_jsons: List[Dict[str, Any]], options: Dict[str, Any]
    ) -> "requests.Response":
        """
        Insert automation documents into the automation collection from a user supplied JSON document.
        https://apidocs.itential.com/2020.2/api/app-automation_catalog/importAutomations/
        :param client: Itential state object
        :param automation_jsons: List of json representations of automation catalogs
        :param options:
        :return: requests.Response: Status of automation import operation
        """
        body = {"automations": automation_jsons, "options": options}
        return client.call(method="POST", url=f"{client.url}/automation_catalog/automations/import", json=body)

    @staticmethod
    def migrate_agenda_jobs(client: "Itential", agenda_jobs: List[str]) -> "requests.Response":
        """
        Takes an existing Agenda Job data structure and converts it to an Automation
        https://apidocs.itential.com/2020.2/api/app-automation_catalog/migrateAgendaJobs/
        :param client: Itential state object
        :param agenda_jobs: List of Unique ids automations
        :return: requests.Response: The result of the migration process
        """
        body = {"agendaJobs": agenda_jobs}
        return client.call(method="POST", url=f"{client.url}/automation_catalog/automations/migration", json=body)

    @staticmethod
    def run_automation(client: "Itential", automation_id: str, options: Dict[str, Any]) -> "requests.Response":
        """
        Single run of an automation outside its scheduled runs, Requires a workflow to be attached to the automation.
        https://apidocs.itential.com/2020.2/api/app-automation_catalog/runAutomation/
        :param client: Itential state object
        :param automation_id: Unique id of the automation
        :param options: Workflow based properties (Mostly undocumented)
        :return: requests.Response: {"status": "str", "message": {}, "workflowResponse": {}}
        """
        body = {"options": options}
        return client.call(
            method="POST", url=f"{client.url}/automation_catalog/automations/{automation_id}/run", json=body
        )

    @staticmethod
    def update_automation(client: "Itential", automation_id: str, options: Dict[str, Any]) -> "requests.Response":
        """
        Updates an automation's attributes (including scheduling data).
        https://apidocs.itential.com/2020.2/api/app-automation_catalog/updateAutomation/
        :param client: Itential state object
        :param automation_id: Unique id of the automation
        :param options: Object containing the fields to be updated
            ex: {"options": {
                    "workflowId": "8e3695fe-c5bf-4286-ae83-186b3fea1c1a",
                    "formId": "74cd8ca4367d4a9617c5c849",
                    "gbac": {
                      "write": [
                        "c47584c0cdb569d258a1e48c",
                        "f8ff16a4d36a61468b6ed5bc"
                      ],
                      "read": [
                        "2f829cfa07d897e682212871",
                        "2cd984705c011c5172458ac4"
                      ]
                    },
                    "nextRunAt": "2019-11-25T22:51:39.201Z",
                    "repeatInterval": null
                  }
                }
        :return: requests.Response: Updated document of requested automation
        """
        body = {"options": options}
        return client.call(method="PUT", url=f"{client.url}/automation_catalog/automations/{automation_id}", json=body)



"""
Link to Itential Docs: https://apidocs.itential.com/2020.2/api/app-nso_manager/

Implemented  Doc String  Tests
    [ ]         [ ]       [ ]   addDeviceForm
    [ ]         [ ]       [ ]   addLockItem
    [ ]         [ ]       [ ]   applyTemplates
    [ ]         [ ]       [ ]   deleteQueueItem
    [ ]         [ ]       [ ]   getAllAuthGroups
    [ ]         [ ]       [ ]   getAllNEDs
    [ ]         [ ]       [ ]   getAuthGroups
    [ ]         [ ]       [ ]   getCommitQueueDeep
    [ ]         [ ]       [ ]   getDevicesFiltered
    [ ]         [ ]       [ ]   getDevicesFilteredForAdapter
    [ ]         [ ]       [ ]   getNEDs
    [ ]         [ ]       [ ]   getQueuedDevices
    [ ]         [ ]       [ ]   getQueueItemDetails
    [ ]         [ ]       [ ]   isAlive
    [ ]         [ ]       [ ]   lockQueueItem
    [ ]         [ ]       [ ]   pruneDevicesAllItems
    [ ]         [ ]       [ ]   pruneDevicesFromItem
    [ ]         [ ]       [ ]   restAction
    [ ]         [ ]       [ ]   restQuery
    [ ]         [ ]       [ ]   runAction
    [ ]         [ ]       [ ]   runCommand
    [ ]         [ ]       [ ]   runCommands
    [ ]         [ ]       [ ]   SetDeviceConfiguration
    [ ]         [ ]       [ ]   setItemNacmGroup
    [ ]         [ ]       [ ]   setLeaf
    [ ]         [ ]       [ ]   unlockQueueItem
    [ ]         [ ]       [ ]   verifyConfig
"""



import logging
from typing import TYPE_CHECKING

import requests

if TYPE_CHECKING:
    from itential.core import Itential

log = logging.getLogger(__name__)


def index_status(client: "Itential", index: str) -> requests.Response:
    """
    Verify that a collection is properly indexed.
    https://apidocs.itential.com/2021.1/admin/Itential%20Automation%20Platform/Configuration/#indexing-apis-and-seeding
    All indexes can be found just above at this link:
    https://apidocs.itential.com/2021.1/admin/Itential%20Automation%20Platform/Configuration/#database-indexes
    """
    return client.call(method="GET", url=f"{client.url}/indexes/{index}/status")


def index_repair(client: "Itential", index: str) -> requests.Response:
    """
    Creates/Repairs all indexes on the server.
    More information can be found here:
    https://apidocs.itential.com/2021.1/admin/Itential%20Automation%20Platform/Configuration/#database-indexes
    """
    return client.call(method="GET", url=f"{client.url}/indexes/{index}")



"""
Link to Itential Docs: https://apidocs.itential.com/2020.2/api/app-configuration_manager/

Implemented  Doc String  Tests
    [ ]         [ ]       [ ]   adapterProxy
    [ ]         [ ]       [ ]   addDevicesToGroup
    [ ]         [ ]       [ ]   addDevicesToNode
    [ ]         [ ]       [ ]   addTasksToNode
    [ ]         [ ]       [ ]   advancedAutoRemediation
    [ ]         [ ]       [ ]   backUpDevice
    [ ]         [ ]       [ ]   buildSpecLines
    [ ]         [ ]       [ ]   createConfigSpec
    [ ]         [ ]       [ ]   createDeviceGroup
    [ ]         [ ]       [ ]   createGoldenConfigNode
    [ ]         [ ]       [ ]   createGoldenConfigTree
    [ ]         [ ]       [ ]   createGoldenConfigTreeVersion
    [ ]         [ ]       [ ]   createJSONSpec
    [ ]         [ ]       [ ]   createTaskInstance
    [ ]         [ ]       [ ]   deleteDeviceBackups
    [ ]         [ ]       [ ]   deleteDeviceGroups
    [ ]         [ ]       [ ]   deleteDeviceGroupsByName
    [ ]         [ ]       [ ]   deleteGoldenConfigNode
    [ ]         [ ]       [ ]   deleteGoldenConfigTree
    [ ]         [ ]       [ ]   deleteGoldenConfigTrees
    [ ]         [ ]       [ ]   deleteGoldenConfigTreeVersion
    [ ]         [ ]       [ ]   deletePins
    [ ]         [ ]       [ ]   deleteTaskInstances
    [ ]         [ ]       [ ]   deleteVariables
    [ ]         [ ]       [ ]   DiffViewer
    [ ]         [ ]       [ ]   exportGoldenConfigTree
    [ ]         [ ]       [ ]   getAdapterTask
    [ ]         [ ]       [ ]   getAdapterTasks
    [ ]         [ ]       [ ]   getBackups
    [ ]         [ ]       [ ]   getComplianceReportDetail
    [ ]         [ ]       [ ]   getComplianceReportDeviceHistory
    [ ]         [ ]       [ ]   getComplianceReportNodeSummary
    [ ]         [ ]       [ ]   getComplianceReportsByBatch
    [ ]         [ ]       [ ]   getComplianceReportsDetail
    [ ]         [ ]       [ ]   getComplianceReportTaskHistory
    [ ]         [ ]       [ ]   getComplianceReportTreeSummary
    [ ]         [ ]       [ ]   getConfigSpec
    [ ]         [ ]       [ ]   getConfigTemplate
    [ ]         [ ]       [ ]   getDevice
    [ ]         [ ]       [ ]   getDeviceBackupById
    [ ]         [ ]       [ ]   getDeviceConfig
    [ ]         [ ]       [ ]   getDeviceGroupById
    [ ]         [ ]       [ ]   getDeviceGroupByName
    [ ]         [ ]       [ ]   getDeviceGroups
    [ ]         [ ]       [ ]   getDevicesFiltered
    [ ]         [ ]       [ ]   getGoldenConfigTree
    [ ]         [ ]       [ ]   getGoldenConfigTrees
    [ ]         [ ]       [ ]   getGoldenConfigTreeVersion
    [ ]         [ ]       [ ]   getJSONComplianceReportDetail
    [ ]         [ ]       [ ]   getJSONComplianceReportsByBatch
    [ ]         [ ]       [ ]   getJSONSpec
    [ ]         [ ]       [ ]   getJSONSpecWithInheritance
    [ ]         [ ]       [ ]   getPins
    [ ]         [ ]       [ ]   getTaskInstances
    [ ]         [ ]       [ ]   getTopIssues
    [ ]         [ ]       [ ]   getTopIssuesJson
    [ ]         [ ]       [ ]   getTreesForDevice
    [ ]         [ ]       [ ]   gradeComplianceReport
    [ ]         [ ]       [ ]   gradeComplianceReports
    [ ]         [ ]       [ ]   gradeDeviceComplianceHistory
    [ ]         [ ]       [ ]   gradeTaskComplianceHistory
    [ ]         [ ]       [ ]   handlePin
    [ ]         [ ]       [ ]   importBackup
    [ ]         [ ]       [ ]   importGoldenConfigTree
    [ ]         [ ]       [ ]   importGroup
    [ ]         [ ]       [ ]   isAlive
    [ ]         [ ]       [ ]   lookupDiff
    [ ]         [ ]       [ ]   ManualRemediation
    [ ]         [ ]       [ ]   ManualRemediationResults
    [ ]         [ ]       [ ]   patchDeviceConfiguration
    [ ]         [ ]       [ ]   removeDevicesFromGroup
    [ ]         [ ]       [ ]   removeDevicesFromNode
    [ ]         [ ]       [ ]   removeTasksFromNode
    [ ]         [ ]       [ ]   renderJinja2
    [ ]         [ ]       [ ]   runAdapterTask
    [ ]         [ ]       [ ]   runAutoRemediation
    [ ]         [ ]       [ ]   runCompliance
    [ ]         [ ]       [ ]   runComplianceForDevice
    [ ]         [ ]       [ ]   runComplianceForNode
    [ ]         [ ]       [ ]   runComplianceForTree
    [ ]         [ ]       [ ]   runTaskInstance
    [ ]         [ ]       [ ]   searchDeviceGroups
    [ ]         [ ]       [ ]   searchGroups
    [ ]         [ ]       [ ]   translateConfigSpec
    [ ]         [ ]       [ ]   updateConfigSpec
    [ ]         [ ]       [ ]   updateDeviceBackupById
    [ ]         [ ]       [ ]   updateDeviceGroups
    [ ]         [ ]       [ ]   updateGoldenConfigNode
    [ ]         [ ]       [ ]   updateGoldenConfigTree
    [ ]         [ ]       [ ]   updateGoldenConfigTreeVersion
    [ ]         [ ]       [ ]   updateJSONSpec
    [ ]         [ ]       [ ]   updateTaskInstance
"""



"""
Link to Itential Docs: https://apidocs.itential.com/2020.2/api/app-operations_manager/

Implemented  Doc String  Tests
    [ ]         [ ]       [ ]   cloneAutomation
    [ ]         [ ]       [ ]   createAutomation
    [ ]         [ ]       [ ]   createTrigger
    [ ]         [ ]       [ ]   deleteAutomation
    [ ]         [ ]       [ ]   deleteTrigger
    [ ]         [ ]       [ ]   deleteTriggersByActionId
    [ ]         [ ]       [ ]   exportAutomation
    [ ]         [ ]       [ ]   exportTrigger
    [ ]         [ ]       [ ]   getAutomation
    [ ]         [ ]       [ ]   getAutomations
    [ ]         [ ]       [ ]   getEventDefinition
    [ ]         [ ]       [ ]   getEvents
    [ ]         [ ]       [ ]   getTrigger
    [ ]         [ ]       [ ]   getTriggers
    [ ]         [ ]       [ ]   importAutomations
    [ ]         [ ]       [ ]   importTriggers
    [ ]         [ ]       [ ]   runEndpointTriggerWithPost
    [ ]         [ ]       [ ]   runManualTrigger
    [ ]         [ ]       [ ]   updateAutomation
    [ ]         [ ]       [ ]   updateTrigger

"""



"""
Link to Itential Docs: https://apidocs.itential.com/2020.2/api/app-template_builder/

Implemented  Doc String  Tests
    [ ]         [ ]       [ ]   applyTemplates
    [ ]         [ ]       [ ]   parseTemplate
    [ ]         [ ]       [ ]   renderJinjaTemplate
"""



"""
Link to Itential Docs: https://apidocs.itential.com/2020.2/api/app-ag_manager/

Implemented  Doc String  Tests
    [x]         [x]       [ ]   discoverModules
    [x]         [x]       [ ]   undiscoverAll
    [x]         [x]       [ ]   undiscoverModules
"""
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from itential.core import Itential
    import requests


class AppAgManager:
    """https://apidocs.itential.com/2020.2/api/app-ag_manager/"""

    @staticmethod
    def discover_modules(client: "Itential", adapter_id: str) -> "requests.Response":
        """
        Discovers all actions from a specified IAG adapter.
        This will restart this app and display a corresponding log message.
        https://apidocs.itential.com/2020.2/api/app-ag_manager/discoverModules/
        :param client: Itential state object
        :param adapter_id: Automation Gateway adapter ID
        :return: requests.Response: A pronghorn.json object.
        """
        return client.call(method="GET", url=f"{client.url}/ag-manager/actions/{adapter_id}")

    @staticmethod
    def undiscover_all(client: "Itential") -> "requests.Response":
        """
        Discovers all actions from all IAG adapters.
        This will restart this app and display a corresponding log message.
        https://apidocs.itential.com/2020.2/api/app-ag_manager/undiscoverAll/
        :param client: Itential state object
        :return: requests.Response: A pronghorn.json object.
        """
        return client.call(method="DELETE", url=f"{client.url}/ag-manager/actions")

    @staticmethod
    def undiscover_modules(client: "Itential", adapter_id: str) -> "requests.Response":
        """
        Removes all actions from a specified IAG adapter.
        This will restart this app and display a corresponding log message.
        https://apidocs.itential.com/2020.2/api/app-ag_manager/undiscoverModules/
        :param client: Itential state object
        :param adapter_id: Automation Gateway adapter ID
        :return: requests.Response: A pronghorn.json object.
        """
        return client.call(method="DELETE", url=f"{client.url}/ag-manager/actions/{adapter_id}")



"""
Link to Itential Docs: https://apidocs.itential.com/2020.2/api/adapter-automation_gateway/

Implemented  Doc String  Tests
    [-]         [-]       [-]   addDeviceToDeviceGroup
    [-]         [-]       [-]   applyTerraform
    [-]         [-]       [-]   connectDevice
    [-]         [-]       [-]   createDevice
    [-]         [-]       [-]   createDeviceGroup
    [-]         [-]       [-]   createDeviceRaw
    [-]         [-]       [-]   deleteDevice
    [-]         [-]       [-]   deleteDeviceFromDeviceGroup
    [-]         [-]       [-]   deleteDeviceGroup
    [-]         [-]       [-]   destroyTerraform
    [-]         [-]       [-]   dryRunPlaybook
    [-]         [-]       [-]   getCollectionModule
    [-]         [-]       [-]   getCollectionModules
    [-]         [-]       [-]   getCollectionRole
    [-]         [-]       [-]   getCollectionRoles
    [-]         [-]       [-]   getCollections
    [-]         [-]       [-]   getConfig
    [-]         [-]       [-]   getDevice
    [-]         [-]       [-]   getDeviceDetailsRaw
    [-]         [-]       [-]   getDeviceGroup
    [-]         [-]       [-]   getDeviceGroups
    [-]         [-]       [-]   getDeviceHealth
    [-]         [-]       [-]   getDeviceInfo
    [-]         [-]       [-]   getDevices
    [-]         [-]       [-]   getDevicesDetailsRaw
    [-]         [-]       [-]   getDevicesFiltered
    [-]         [-]       [-]   getModule
    [-]         [-]       [-]   getModules
    [-]         [-]       [-]   getNornirModule
    [-]         [-]       [-]   getNornirModules
    [-]         [-]       [-]   getPlaybook
    [-]         [-]       [-]   getPlaybooks
    [-]         [-]       [-]   getPronghorn
    [-]         [-]       [-]   getRole
    [-]         [-]       [-]   getRoles
    [-]         [-]       [-]   getScript
    [-]         [-]       [-]   getScripts
    [-]         [-]       [-]   getTerraform
    [-]         [-]       [-]   getTerraforms
    [-]         [-]       [-]   initTerraform
    [-]         [-]       [-]   isAlive
    [x]         [x]       [ ]   netmikoSendCommand
    [x]         [x]       [ ]   netmikoSendConfig
    [-]         [-]       [-]   planTerraform
    [-]         [-]       [-]   refreshCollections
    [-]         [-]       [-]   refreshInventory
    [-]         [-]       [-]   refreshModules
    [-]         [-]       [-]   refreshNornirModules
    [-]         [-]       [-]   refreshPlaybooks
    [-]         [-]       [-]   refreshRoles
    [-]         [-]       [-]   refreshScripts
    [-]         [-]       [-]   restoreConfig
    [-]         [-]       [-]   runCollectionModule
    [-]         [-]       [-]   runCollectionRole
    [-]         [-]       [-]   runCommand
    [-]         [-]       [-]   runModule
    [-]         [-]       [-]   runNornirModule
    [-]         [-]       [-]   runPlaybook
    [-]         [-]       [-]   runRole
    [-]         [-]       [-]   runScript
    [-]         [-]       [-]   runScriptEnv
    [-]         [-]       [-]   setConfig
    [-]         [-]       [-]   updateDevice
    [-]         [-]       [-]   updateDeviceGroup
"""
from typing import TYPE_CHECKING, Dict, List, Union

import requests

if TYPE_CHECKING:
    from itential.core import Itential


class AdapterAutomationGateway:
    """https://apidocs.itential.com/2020.2/api/adapter-automation_gateway/"""

    @staticmethod
    def netmiko_send_command(client: "Itential", host: str, commands: List[str],
                             connection_options: Dict[str, str]) -> requests.Response:
        """
        Wrapper of send_command of netmiko.

        https://apidocs.itential.com/2020.2/api/adapter-automation_gateway/netmikoSendCommand/

        :param client: The Itential state object
        :param host: Device TID/Name
        :param commands: A List of commands to run.
        :param connection_options: Connection and Authentication information.
         Keys: "device_type", "port", "username", "password".
         See Netmiko documentation for more detail.
        :return: A JSON Object containing status, code and the result
        """
        body = {
            "netmikoSendCommandParameters": {
                "host": host,
                "commands": commands,
                "connection_options": connection_options
            }
        }
        return client.call(method="POST", url=f"{client.url}/automationgateway/netmiko/send_command", json=body)

    @staticmethod
    def netmiko_send_config(client: "Itential", host: str, config_commands: List[str],
                            connection_options: Dict[str, Union[int, str]]) -> requests.Response:
        """
        Wrapper of send_config_set of netmiko.

        https://apidocs.itential.com/2020.2/api/adapter-automation_gateway/netmikoSendConfig/

        :param client: The Itential state object
        :param host: Device TID/Name
        :param config_commands: A List of config commands to run.
        :param connection_options: Connection and Authentication information.
         Keys: "device_type", "port", "username", "password".
         See Netmiko documentation for more detail.
        :return: A JSON Object containing status, code and the result
        """
        body = {
            "netmikoSendConfigParameters": {
                "host": host,
                "config_commands": config_commands,
                "connection_options": connection_options
            }
        }
        return client.call(method="POST", url=f"{client.url}/automationgateway/netmiko/send_config", json=body)



"""
Link to Itential Docs: https://apidocs.itential.com/2020.2/api/app-workflow_builder/

Implemented  Doc String  Tests
    [x]         [x]       [ ]   createWorkflowGroupEntry
    [x]         [x]       [ ]   deleteWorkflow
    [x]         [x]       [ ]   deleteWorkflowGroups
    [x]         [x]       [ ]   exportWorkflow
    [x]         [x]       [ ]   getSchemas
    [x]         [x]       [ ]   getTaskDetails
    [x]         [x]       [ ]   getTasksList
    [x]         [x]       [ ]   importWorkflow
    [x]         [x]       [ ]   listWorkflowGroups
    [x]         [x]       [ ]   removeWorkflowGroup
    [x]         [x]       [ ]   renameWorkflow
    [x]         [x]       [ ]   replaceWorkflowGroups
    [x]         [x]       [ ]   saveWorkflow
"""
from typing import TYPE_CHECKING, Dict, List, Any

if TYPE_CHECKING:
    import requests
    from itential.core import Itential


class AppWorkflowBuilder:
    """
    https://apidocs.itential.com/2020.2/api/app-workflow_builder/
    """

    @staticmethod
    def delete_workflow(client: "Itential", name: str) -> "requests.Response":
        """
        Deletes a single workflow of the given name.
        https://apidocs.itential.com/2020.2/api/app-workflow_builder/deleteWorkflow/
        :param client: Itential state object
        :param name: Name of the workflow to be deleted.
        :return: requests.Response object
                 A successful delete returns the deleted workflow json.
                 A failed delete returns a 500: "TypeError: Cannot read property '_id' of null"
        """
        return client.call(method="DELETE", url=f"{client.url}/workflow_builder/workflows/delete/{name}")

    @staticmethod
    def import_workflow(client: "Itential", data: Dict[str, Any]) -> "requests.Response":
        """
        Imports a single workflow.
        https://apidocs.itential.com/2020.2/api/app-workflow_builder/importWorkflow/
        :param client: Itential state object
        :param data: The workflow json object.
        :return: requests.Response object. Successful response ex: {'n': 1, 'ok': 1, 'name': 'automation_name'}
        """
        data = {"workflow": data, "options": {}}
        return client.call(method="POST", url=f"{client.url}/workflow_builder/import", json=data)

    @staticmethod
    def create_workflow_group_entry(client: "Itential", workflow_name: str, group_name: str) -> "requests.Response":
        """
        Add Group to a Workflow.
        https://apidocs.itential.com/2020.2/api/app-workflow_builder/createWorkflowGroupEntry/
        :param client: Itential state object
        :param workflow_name: Name of the workflow to be assigned to the group
        :param group_name: < Unclear atm. Docs show an id, but the description says "name".
        :return: requests.Response
                                    500 error: "Invalid group: TestGroupName"
        """
        data = {"group": group_name}
        return client.call(
            method="POST", url=f"{client.url}/workflow_builder/workflows/{workflow_name}/groups", json=data
        )

    @staticmethod
    def delete_workflow_groups(client: "Itential", workflow_name: str, group_name: str) -> "requests.Response":
        """
        Delete all Groups for a Workflow
        https://apidocs.itential.com/2020.2/api/app-workflow_builder/deleteWorkflowGroups/ (Docs are incomplete)
        :param client: Itential state object
        :param workflow_name: Name of the workflow to delete
        :param group_name: < unclear atm > Itential docs aren't complete for this version.
        :return: requests.Response  200: boolean
                                    500: "Could not find workflow: TestWorkflow". The group_name doesn't seem to matter.
        """
        data = {'groups': group_name}
        return client.call(
            method="DELETE", url=f"{client.url}/workflow_builder/workflows/{workflow_name}/groups", json=data
        )

    @staticmethod
    def export_workflow(client: "Itential", workflow_id: str = '', workflow_name: str = '') -> "requests.Response":
        """
        Returns the workflow json (Does not include a "_id" key at the top level)
        https://apidocs.itential.com/2020.2/api/app-workflow_builder/exportWorkflow/
        :param client: Itential state object
        :param workflow_id: ID of the workflow to export
        :param workflow_name: Name of the workflow to export.
        :return: requests.Response  200: Returns the full workflow json.
                Wrong workflow name 500: "Could not find matching workflow"
                Wrong workflow ID   500: "Could not find matching workflow"
        """
        data = {'options': {"type": "automation"}}  # "type" is optional

        if workflow_id:
            data["options"]["_id"] = workflow_id

        elif workflow_name:
            data["options"]["name"] = workflow_name

        else:
            raise AttributeError("Must have one of 'workflow_id' or 'workflow_name'.")

        return client.call(method="POST", url=f"{client.url}/workflow_builder/export", json=data)

    @staticmethod
    def get_schemas(client: "Itential", workflow_object: Dict[str, Any]) -> "requests.Response":
        """
        Calculate incoming/outgoing schemas for the workflow
        https://apidocs.itential.com/2020.2/api/app-workflow_builder/getSchemas/
        :param client: Itential state object
        :param workflow_object: A workflow json object.
        :return: requests.Response  200: {"inputSchema": {...}, "outputSchema": {...}}
        """
        data = {"workflow": workflow_object}
        return client.call(method="POST", url=f"{client.url}/workflow_builder/workflows/schemas", json=data)

    @staticmethod
    def get_task_details(client: "Itential", app_name: str, task_id: str) -> "requests.Response":
        """
        Get Task Details
        https://apidocs.itential.com/2020.2/api/app-workflow_builder/getTaskDetails/
        :param client: Itential state object
        :param app_name: Application Name (Export field in model)
        :param task_id: Task ID (Hexadecimal). < Unclear what this variable is >
        :return: requests.Response  200:
        """
        return client.call(method="GET", url=f"{client.url}/getTaskDetails/{app_name}/{task_id}")

    @staticmethod
    def get_tasks_list(client: "Itential") -> "requests.Response":
        """
        Get all Tasks.
        https://apidocs.itential.com/2020.2/api/app-workflow_builder/getTasksList/
        :param client: Itential state object
        :return: requests.Response
        """
        return client.call(method="GET", url=f"{client.url}/workflow_builder/tasks/list")

    @staticmethod
    def list_workflow_groups(client: "Itential", workflow_name: str) -> "requests.Response":
        """
        List the groups that have access to a Workflow
        https://apidocs.itential.com/2020.2/api/app-workflow_builder/listWorkflowGroups/
        :param client: Itential state object
        :param workflow_name: Name of the workflow to query
        :return: requests.Response
        """
        return client.call(method="GET", url=f"{client.url}/workflow_builder/workflows/{workflow_name}/groups")

    @staticmethod
    def remove_workflow_group(client: "Itential", workflow_name: str, group_name: str) -> "requests.Response":
        """
        Remove a group from the list of authorized groups for a Workflow
        https://apidocs.itential.com/2020.2/api/app-workflow_builder/removeWorkflowGroup/
        :param client: Itential state object
        :param workflow_name: Name of workflow to remove group from
        :param group_name: Name of group to remove from workflow.
        :return: requests.Response
        """
        return client.call(
            method="DELETE", url=f"{client.url}/workflow_builder/workflows/{workflow_name}/groups/{group_name}"
        )

    @staticmethod
    def rename_workflow(client: "Itential", workflow_object: Dict[str, Any], new_name: str) -> "requests.Response":
        """
        Rename a Workflow in the database
        https://apidocs.itential.com/2020.2/api/app-workflow_builder/renameWorkflow/
        :param client: Itential state object
        :param workflow_object: Current workflow object. Minimum requirements are {"_id": "", "name": ""}
        :param new_name: New workflow name.
        :return: requests.Response
        """
        data = {"workflow": workflow_object, "newName": new_name}
        return client.call(method="POST", url=f"{client.url}/workflow_builder/workflows/rename", json=data)

    @staticmethod
    def replace_workflow_groups(client: "Itential", workflow_name: str, group_list: List[str]) -> "requests.Response":
        """
        Overwrite the list of groups that have access to a Workflow
        https://apidocs.itential.com/2020.2/api/app-workflow_builder/replaceWorkflowGroups/
        :param client: Itential state object
        :param workflow_name: Name of the workflow
        :param group_list: List of group IDs (Hexadecimal)
        :return: requests.Response
        """
        data = {"groups": group_list}
        return client.call(
            method="PUT", url=f"{client.url}/workflow_builder/workflows/{workflow_name}/groups/", json=data
        )

    @staticmethod
    def save_workflow(client: "Itential", workflow_obj: Dict[str, Any]) -> "requests.Response":
        """
        Add a Workflow to the database (Basically the save button)
        https://apidocs.itential.com/2020.2/api/app-workflow_builder/saveWorkflow/
        :param client: Itential state object
        :param workflow_obj: Full workflow json
        :return: requests.Response
        """
        data = {"workflow": workflow_obj}
        return client.call(method="POST", url=f"{client.url}/workflow_builder/workflows/save", json=data)



"""
Link to Itential Docs: https://apidocs.itential.com/2020.2/api/app-workflow_engine/

Implemented  Doc String  Tests  Fix Method Tooltip
    [x]         [x]       [ ]         [ ]   activate
    [-]         [-]       [-]         [ ]   addDuration
    [x]         [x]       [ ]         [ ]   addWatchers
    [-]         [-]       [-]         [ ]   arrayConcat
    [-]         [-]       [-]         [ ]   arrayIncludes
    [-]         [-]       [-]         [ ]   arrayIndexOf
    [-]         [-]       [-]         [ ]   arrayLastIndexOf
    [-]         [-]       [-]         [ ]   arrayLength
    [-]         [-]       [-]         [ ]   arrayPop
    [-]         [-]       [-]         [ ]   arrayPush
    [-]         [-]       [-]         [ ]   arrayShift
    [-]         [-]       [-]         [ ]   arraySlice
    [-]         [-]       [-]         [ ]   arrayToLocaleString
    [-]         [-]       [-]         [ ]   arrayToString
    [-]         [-]       [-]         [ ]   assign
    [-]         [-]       [-]         [ ]   calculateTimeDiff
    [x]         [x]       [ ]         [ ]   cancelJob
    [-]         [-]       [-]         [ ]   charAt
    [-]         [-]       [-]         [ ]   charCodeAt
    [x]         [x]       [ ]         [ ]   checkWorkflowForJobVariables
    [-]         [-]       [-]         [ ]   childJob
    [x]         [x]       [-]         [ ]   claimTask
    [-]         [-]       [-]         [ ]   codePointAt
    [-]         [-]       [-]         [ ]   convertEpochToObject
    [-]         [-]       [-]         [ ]   convertTimeFormat
    [-]         [-]       [-]         [ ]   convertTimeToEpoch
    [-]         [-]       [-]         [ ]   convertTimezone
    [-]         [-]       [-]         [ ]   copyWithin
    [x]         [x]       [ ]         [ ]   createJobGroupEntry
    [x]         [x]       [ ]         [ ]   deactivate
    [-]         [-]       [-]         [ ]   decision
    [-]         [-]       [-]         [ ]   deepmerge
    [-]         [-]       [-]         [ ]   delay
    [x]         [x]       [ ]         [ ]   deleteJobGroups
    [x]         [x]       [ ]         [ ]   diffToHTML
    [-]         [-]       [-]         [ ]   endsWith
    [-]         [-]       [-]         [ ]   ErrorHandling
    [-]         [-]       [-]         [ ]   evaluation
    [-]         [-]       [-]         [ ]   eventListenerJob
    [-]         [-]       [-]         [ ]   extractField
    [-]         [-]       [-]         [ ]   fill
    [x]         [x]       [ ]         [ ]   find
    [x]         [x]       [ ]         [ ]   findForwardPaths
    [x]         [x]       [ ]         [ ]   finishManualTask
    [x]         [x]       [ ]         [ ]   fixJob
    [-]         [-]       [-]         [-]   FlattenJSONFormInput
    [-]         [-]       [-]         [-]   forEach
    [-]         [-]       [-]         [-]   getAllLoopTasks
    [x]         [x]       [ ]         [x]   getAssociatedJobs
    [x]         [x]       [ ]         [x]   getEntireJob
    [x]         [x]       [ ]         [x]   getJob
    [x]         [x]       [ ]         [x]   getJobDetails
    [x]         [x]       [ ]         [x]   getJobFromTaskQuery
    [x]         [x]       [ ]         [x]   getJobList
    [x]         [x]       [ ]         [ ]   getJobOutput
    [x]         [x]       [ ]         [x]   getJobShallow
    [x]         [x]       [ ]         [x]   getJobVisualizationData
    [x]         [x]       [ ]         [x]   getManualTaskController
    [x]         [x]       [ ]         [x]   getTask
    [x]         [x]       [ ]         [x]   getTaskDetails
    [x]         [x]       [ ]         [x]   getTaskStatuses
    [-]         [-]       [-]         [-]   getTime
    [x]         [x]       [ ]         [x]   getWorkflowsDetailedByName
    [x]         [x]       [ ]         [x]   isActive
    [-]         [-]       [-]         [-]   isArray
    [-]         [-]       [-]         [-]   join
    [-]         [-]       [-]         [-]   keys
    [x]         [x]       [ ]         [x]   listJobGroups
    [-]         [-]       [-]         [-]   localeCompare
    [-]         [-]       [-]         [-]   makeData
    [-]         [-]       [-]         [-]   map
    [-]         [-]       [-]         [-]   match
    [-]         [-]       [-]         [-]   merge
    [-]         [-]       [-]         [-]   modify
    [-]         [-]       [-]         [-]   newVariable
    [-]         [-]       [-]         [-]   normalize
    [-]         [-]       [-]         [-]   numberToString
    [-]         [-]       [-]         [-]   objectHasOwnProperty
    [-]         [-]       [-]         [-]   objectToString
    [-]         [-]       [-]         [-]   padEnd
    [-]         [-]       [-]         [-]   padStart
    [-]         [-]       [-]         [-]   parse
    [-]         [-]       [-]         [-]   parseInt
    [x]         [x]       [ ]         [x]   pauseJob
    [-]         [-]       [-]         [-]   pop
    [x]         [x]       [ ]         [x]   prepareMetricsLogs
    [-]         [-]       [-]         [-]   push
    [-]         [-]       [-]         [-]   query
    [x]         [x]       [ ]         [x]   queryJobs
    [x]         [x]       [ ]         [x]   queryTasksBrief
    [x]         [x]       [ ]         [x]   releaseTask
    [x]         [x]       [ ]         [x]   removeJobGroup
    [-]         [-]       [-]         [-]   repeat
    [-]         [-]       [-]         [-]   replace
    [x]         [x]       [ ]         [x]   replaceJobGroups
    [-]         [-]       [-]         [-]   restCall
    [x]         [x]       [ ]         [x]   resumeJob
    [-]         [-]       [-]         [-]   reverse
    [x]         [x]       [ ]         [x]   revertToTask
    [x]         [x]       [ ]         [x]   runEvaluationGroup
    [x]         [x]       [ ]         [x]   runEvaluationGroups
    [x]         [x]       [ ]         [x]   runValidation
    [-]         [-]       [-]         [-]   search
    [x]         [x]       [ ]         [x]   searchJobs
    [x]         [x]       [ ]         [x]   searchTasks
    [x]         [x]       [ ]         [x]   searchWorkflows
    [-]         [-]       [-]         [-]   setObjectKey
    [-]         [-]       [-]         [-]   shift
    [-]         [-]       [-]         [-]   sort
    [-]         [-]       [-]         [-]   split
    [x]         [x]       [ ]         [x]   startJobWithOptions
    [-]         [-]       [-]         [-]   startsWith
    [-]         [-]       [-]         [-]   stringConcat
    [-]         [-]       [-]         [-]   stringIncludes
    [-]         [-]       [-]         [-]   stringIndexOf
    [-]         [-]       [-]         [-]   stringLastIndexOf
    [-]         [-]       [-]         [-]   stringLength
    [-]         [-]       [-]         [-]   stringSlice
    [-]         [-]       [-]         [-]   stringValueOf
    [-]         [-]       [-]         [-]   stub
    [-]         [-]       [-]         [-]   substring
    [-]         [-]       [-]         [-]   toLocaleLowerCase
    [-]         [-]       [-]         [-]   toLocaleUpperCase
    [-]         [-]       [-]         [-]   toLowerCase
    [-]         [-]       [-]         [-]   toUpperCase
    [-]         [-]       [-]         [-]   transformation
    [-]         [-]       [-]         [-]   trim
    [-]         [-]       [-]         [-]   trimEnd
    [-]         [-]       [-]         [-]   trimStart
    [-]         [-]       [-]         [-]   unshift
    [x]         [x]       [ ]         [x]   unwatchJob
    [-]         [-]       [-]         [-]   updateJobDescription
    [x]         [x]       [ ]         [x]   validateAllLoops
    [-]         [-]       [-]         [-]   values
    [-]         [-]       [-]         [-]   ViewData
    [-]         [-]       [-]         [-]   ViewDiff
    [x]         [x]       [ ]         [x]   watchJob
"""
import warnings
from typing import TYPE_CHECKING, Optional, Dict, Any, List

import requests

if TYPE_CHECKING:
    from itential.core import Itential


class AppWorkflowEngine:
    """https://apidocs.itential.com/2020.2/api/app-workflow_engine/"""

    @staticmethod
    def activate(client: "Itential", **kwargs: Dict[str, Any]) -> requests.Response:
        """
        Activate Task Worker ( API payload not documented. May not be usable/implemented )
        https://apidocs.itential.com/2020.2/api/app-workflow_engine/activate/
        :param client: The Itential state object
        :param kwargs: Undocumented payload kwargs
        :return: Status flag of activation
        """
        return client.call(method="POST", url=f"{client.url}/workflow_engine/activate", json=kwargs)

    @staticmethod
    def add_watchers(client: "Itential", job_id: str, watchers: List[str]) -> requests.Response:
        """
        Add users to the watchers list of a job by job ID and username.
        https://apidocs.itential.com/2020.2/api/app-workflow_engine/addWatchers/
        :param client: The Itential state object
        :param job_id: ID of the job to watch.
        :param watchers: Users to add to watchers list.
        :return: User IDs added to the watchers list.
        """
        body = {"job_id": job_id, "watchers": watchers}
        return client.call(method="POST", url=f"{client.url}/workflow_engine/job/watchers/watch", json=body)

    @staticmethod
    def cancel_job(client: "Itential", job_id: str) -> requests.Response:
        """
        Cancel an active job by job ID.
        https://apidocs.itential.com/2020.2/api/app-workflow_engine/cancelJob/
        :param client: The Itential state object
        :param job_id: ID of the job to cancel.
        :return: Data from the canceled job.
        """
        body = {"job_id": job_id}
        return client.call(method="POST", url=f"{client.url}/workflow_engine/cancelJob", json=body)

    @staticmethod
    def check_workflow_for_job_variables(client: "Itential", workflow_name: str) -> requests.Response:
        """
        Get job variables of a workflow by workflow name.
        https://apidocs.itential.com/2020.2/api/app-workflow_engine/checkWorkflowForJobVariables/
        :param client: The Itential state object
        :param workflow_name: Name of the workflow.
        :return: Variables of the workflow matching the name.
        """
        return client.call(method="GET", url=f"{client.url}/workflow_engine/workflows/variables/{workflow_name}")

    @staticmethod
    def claim_task(client: "Itential", task_id: str, user: str) -> requests.Response:
        """
        Claim a manual Task
        https://apidocs.itential.com/2020.2/api/app-workflow_engine/claimTask/
        :param client: The Itential state object
        :param task_id: ID of the Task to claim.
        :param user: User id for the user claiming the Task.
        :return: Result of claiming the Task.
        """
        body = {"task_id": task_id, "user": user}
        return client.call(method="POST", url=f"{client.url}/workflow_engine/tasks/claim", json=body)

    @staticmethod
    def create_job_group_entry(client: "Itential", job_id: str, group_id: str) -> requests.Response:
        """
        Add a group to the list of groups for a job.
        https://apidocs.itential.com/2020.2/api/app-workflow_engine/createJobGroupEntry/
        :param client: The Itential state object
        :param job_id: ID of job
        :param group_id: A Group id
        :return: Status of adding a group to a job.
        """
        body = {"group": group_id}
        return client.call(method="POST", url=f"{client.url}/workflow_engine/jobs/{job_id}/groups", json=body)

    @staticmethod
    def deactivate(client: "Itential", **kwargs: Dict[str, Any]) -> requests.Response:
        """
        Deactivate Task Worker ( API payload not documented. May not be usable/implemented )
        https://apidocs.itential.com/2020.2/api/app-workflow_engine/deactivate/
        :param client: The Itential state object
        :param kwargs: Undocumented POST payload
        :return: Status flag of deactivation
        """
        return client.call(method="POST", url=f"{client.url}/workflow_engine/deactivate", json=kwargs)

    @staticmethod
    def delete_job_groups(client: "Itential", job_id: str) -> requests.Response:
        """
        Remove all authorization restriction for a job.
        https://apidocs.itential.com/2020.2/api/app-workflow_engine/deleteJobGroups/
        :param client: The Itential state object
        :param job_id: ID of job
        :return: Status of Delete
        """
        return client.call(method="DELETE", url=f"{client.url}/workflow_engine/jobs/{job_id}/groups")

    @staticmethod
    def diff_to_html(client: "Itential", label1: str, value1: str, label2: str, value2: str) -> requests.Response:
        """
        DEPRECIATED: Difference between two values in HTML response
        https://apidocs.itential.com/2020.2/api/app-workflow_engine/diffToHTML/
        :param client: The Itential state object
        :param label1: Optional label for value1
        :param value1: Any value to diff against value2
        :param label2: Optional label for value2
        :param value2: Any value to diff against value1
        :return: HTML difference between value1 and value2
        """
        warnings.warn("Marked as depreciated by Itential.", PendingDeprecationWarning)
        body = {"label1": label1, "value1": value1, "label2": label2, "value2": value2}
        return client.call(method="POST", url=f"{client.url}/workflow_engine/diffToHTML", json=body)

    @staticmethod
    def find(client: "Itential", query: Dict[str, Any], options: Dict[str, Any]) -> requests.Response:
        """
        Find job Documents based on a query and additional options.
        https://apidocs.itential.com/2020.2/api/app-workflow_engine/find/
        :param client: The Itential state object
        :param query: A MongoDB query object. See docs for an example
        :param options: A MongoDB options object. See docs for an example
        :return: Resulting job Documents that matched query and options.
        """
        body = {"query": query, "options": options}
        return client.call(method="POST", url=f"{client.url}/workflow_engine/jobs/find", json=body)

    @staticmethod
    def find_forward_paths(
        client: "Itential", start_task_id: str, end_task_id: str, workflow_details: Dict[str, Any]
    ) -> requests.Response:
        """
        Find the paths between two Tasks in a Workflow by Task ids and Workflow details.
        https://apidocs.itential.com/2020.2/api/app-workflow_engine/findForwardPaths/
        :param client: The Itential state object
        :param start_task_id: ID of the start Task.
        :param end_task_id: ID of the end Task.
        :param workflow_details: Workflow data which contain the tasks and transitions properties of the workflow.
        :return: All valid paths between the Tasks.
        """
        body = {"start_task": start_task_id, "end_task": end_task_id, "workflow_details": workflow_details}
        return client.call(method="POST", url=f"{client.url}/workflow_engine/findForwardPaths", json=body)

    @staticmethod
    def finish_manual_task(
        client: "Itential", task_id: str, job_id: str, task_data: Dict[str, Any]
    ) -> requests.Response:
        """
        Finishes a manual task with finish state success
        https://apidocs.itential.com/2020.2/api/app-workflow_engine/finishManualTask/
        :param client: The Itential state object
        :param task_id: ID of the Task within that job.
        :param job_id: ID of the Job.
        :param task_data: Data of the completed Task.
        :return: The Current Job Object
        """
        body = {"task_id": task_id, "job_id": job_id, "task_data": task_data}
        return client.call(method="POST", url=f"{client.url}/workflow_engine/finishTask", json=body)

    @staticmethod
    def fix_job(client: "Itential", job_id: str, errored_task: str, revert_task: str) -> requests.Response:
        """
        Revert an errored job to a target task by job ID and task names.
        https://apidocs.itential.com/2020.2/api/app-workflow_engine/fixJob/
        :param client: The Itential state object
        :param job_id: ID of the job to fix.
        :param errored_task: Name of the errored Task.
        :param revert_task: Name of the target Task.
        :return: Job after the revert.
        """
        body = {"job_id": job_id, "errored_task": errored_task, "revert_task": revert_task}
        return client.call(method="POST", url=f"{client.url}/workflow_engine/fixJob", json=body)

    @staticmethod
    def get_all_loop_tasks(client: "Itential", workflow_details: Dict[str, Any]) -> requests.Response:
        """
        Get all looped Tasks in a Workflow by Workflow details.
        https://apidocs.itential.com/2020.2/api/app-workflow_engine/getAllLoopTasks/
        :param client: The Itential state object
        :param workflow_details: Workflow to get looped tasks.
        :return: looped Tasks of the Workflow.
        """
        body = {"workflow_details": workflow_details}
        return client.call(method="POST", url=f"{client.url}/workflow_engine/getAllLoopTasks", json=body)

    @staticmethod
    def start_job_with_options(client: "Itential", workflow_name: str, options: Dict[str, Any]) -> requests.Response:
        """
        Initiate a job of a workflow with options.

        https://apidocs.itential.com/2020.2/api/app-workflow_engine/startJobWithOptions/

        :param client: The Itential state object
        :param workflow_name: Name of the workflow from which to start the job.
        :param options: Parameters to start the workflow with
            Ex: "options": {
                    "description": "str",
                    "variables": {
                        "incoming": {
                        ...
                        }
                    }
                }
        :return: Job which is started.
        """
        body = {"options": options}
        return client.call(
            method="POST", url=f"{client.url}/workflow_engine/startJobWithOptions/{workflow_name}", json=body
        )

    @staticmethod
    def get_associated_jobs(
        client: "Itential",
        filters: Optional[Dict[str, str]] = None,
        sort: Optional[Dict[str, int]] = None,
        limit: int = 10,
        skip: int = 0,
    ) -> requests.Response:
        """
        Search for jobs that the user has touched.

        https://apidocs.itential.com/2020.2/api/app-workflow_engine/getAssociatedJobs/

        :param client: The Itential state object
        :param filters: Ex: {"metrics.user": "admin@pronghorn"} to filter for a specific user.
        :param skip: Number of jobs to offset the result. Useful for pagination.
        :param sort: {"name": -1} or {"name": 1}. "name" field can be any field in the workflow json.
        :param limit: Set to 0 to disable pagination.
        :return: All matching jobs. example json(): {"results": ["_id": int, "name": str],
         "skip": int, "limit": int, "total": int}
        """

        body = {
            "options": {
                # filter: {"key.path": "value"}  Added below
                "page": {"skip": skip, "limit": limit},
                "sort": {"field": "name", "direction": -1},
            }
        }

        if filters:
            body["options"]["filter"] = filters

        if sort:
            body["options"]["sort"] = sort

        return client.call(method="POST", url=f"{client.url}/workflow_engine/getAllLoopTasks", json=body)

    @staticmethod
    def get_entire_job(client: "Itential", job_id: str) -> requests.Response:
        """
        Get entire data of a job by job ID.
        Returns an "exploded" view, where data input into tasks is displayed in a schema view as opposed to the same
        format as it was originally passed in as. Contained about 20% more lines than get_job without containing any
        additional data.

        https://apidocs.itential.com/2020.2/api/app-workflow_engine/getEntireJob/

        :param client: Itential client object. Passed in to all commands
        :param job_id: Returned when creating a job or querying for jobs by workflow name.
         Ex: "ec59ef85fef84e59bf36bd1e"
        :return: requests.Response
        """
        return client.call(method="GET", url=f"{client.url}/workflow_engine/getEntireJob/{job_id}")

    @staticmethod
    def get_job(client: "Itential", job_id: str) -> requests.Response:
        """
        Get a job by job ID.
        The main difference between this endpoint and get_entire_job is that this endpoint seems to have more concise
        input layouts. Showing just the input json as it was passed in, instead of the input json as it is in the schema
        (more of an exploded schema view). About 20% more concise than get_entire_job (limited testing, ~2350 lines from
        get_entire_job down to ~1875 lines from get_job).

        https://apidocs.itential.com/2020.2/api/app-workflow_engine/getJob/

        :param client: Itential client object. Passed in to all commands
        :param job_id: Returned when creating a job or querying for jobs by workflow name.
         Ex: "ec59ef85fef84e59bf36bd1e"
        :return: requests.Response
        """
        return client.call(method="GET", url=f"{client.url}/workflow_engine/getJob/{job_id}")

    @staticmethod
    def get_job_details(client: "Itential", job_id: str) -> requests.Response:
        """
        Get a job by job ID.
        Similar to get_job. Returns a smaller payload than get_job. Incoming json variables are shortened to the
        Itential variables (ex $var.60d8.merged_object) as opposed to the entire json being passed into a task. This
        endpoint also excludes some meta data like: created, created_by, last_updated, last_updated_by, etc
        Response is about 20% lighter than get_job (scaling with task input sizes) (1535 vs 1873 lines for same job)

        https://apidocs.itential.com/2020.2/api/app-workflow_engine/getJobDetails/

        :param client: Itential client object. Passed in to all commands
        :param job_id: Returned when creating a job or querying for jobs by workflow name.
         Ex: "ec59ef85fef84e59bf36bd1e"
        :return: requests.Response
        """
        return client.call(method="GET", url=f"{client.url}/workflow_engine/job/{job_id}/details")

    @staticmethod
    def get_job_from_task_query(client: "Itential", task_query: str, options: Dict[str, Any]) -> requests.Response:
        """
        Search for jobs using a task query.

        https://apidocs.itential.com/2020.2/api/app-workflow_engine/getJobFromTaskQuery/

        :param client: Itential client object. Passed in to all commands
        :param task_query: A mongodb query partial to run against the tasks to get the requested job.
         Examples: {"status": "running"}
        :param options: Sort, filter, and pagination options.
         Examples: {"filter": {"name": "<workflowName>"}, "page": {"skip": 0, "limit": 50}, "sort:
          {"field": "name", "direction": 1}}
        :return: requests.Response. A list of full job objects
        """
        body = {"task_query": task_query, "options": options}
        return client.call(method="POST", url=f"{client.url}/workflow_engine/getJobFromTaskQuery", json=body)

    @staticmethod
    def get_job_list(
        client: "Itential",
        status: str = "active",
        filters: Optional[Dict[str, str]] = None,
        limit: int = 10,
        skip: int = 0,
    ) -> requests.Response:
        """
        Get a list of jobs by status.

        https://apidocs.itential.com/2020.2/api/app-workflow_engine/getJobList/

        :param client: The Itential client state object, required for authentication with the server.
        :param filters: Ex: {"metrics.user": "admin@pronghorn"} to filter for a specific user.
        :param skip: Number of jobs to offset the result. Useful for pagination.
        :param status: Options: "active", "cancelled", "complete", "error", "incomplete", "paused", "running"
        :param limit: Set to 0 to disable pagination.
        :return: requests.Response
        """

        body = {
            "options": {
                "fields": {"tasks": 0},
                # filter: {"key.path": "value"}  Added below
                "page": {"skip": skip, "limit": limit},
                "sort": {"field": "name", "direction": -1},
            }
        }

        if filters:
            body["options"]["filter"] = filters

        return client.call(method="POST", url=f"{client.url}/workflow_engine/getJobList/{status}", json=body)

    @staticmethod
    def get_job_output(client: "Itential", job_id: str) -> requests.Response:
        """
        Get the output of a completed job.

        https://apidocs.itential.com/2020.2/api/app-workflow_engine/getJobOutput/

        :param client: Itential client object. Passed in to all commands
        :param job_id: Returned when creating a job or querying for jobs by workflow name.
         Ex: "ec59ef85fef84e59bf36bd1e"
        :return: requests.Response
        """
        return client.call(method="GET", url=f"{client.url}/workflow_engine/job/{job_id}/output")

    @staticmethod
    def get_job_shallow(client: "Itential", job_id: str) -> requests.Response:
        """
        Get shallow data of a job by job ID. Returns a subset of data returned by omitting tasks' application, tasks'
        incoming arguments, tasks' returned data, and tasks' error information.
        About a 50% smaller payload than get_job (1535 vs 969)

        https://apidocs.itential.com/2020.2/api/app-workflow_engine/getJobShallow/

        :param client: Itential client object. Passed in to all commands
        :param job_id: Returned when creating a job or querying for jobs by workflow name.
         Ex: "ec59ef85fef84e59bf36bd1e"
        :return: requests.Response
        """
        return client.call(method="GET", url=f"{client.url}/workflow_engine/getJobShallow/{job_id}")

    @staticmethod
    def get_job_visualization_data(client: "Itential", job_id: str) -> requests.Response:
        """
        Get a job's visualization data by job ID.

        https://apidocs.itential.com/2020.2/api/app-workflow_engine/getJobVisualizationData/

        :param client: Itential client object. Passed in to all commands
        :param job_id: Returned when creating a job or querying for jobs by workflow name.
         Ex: "ec59ef85fef84e59bf36bd1e"
        :return: requests.Response
        """
        return client.call(method="GET", url=f"{client.url}/workflow_engine/jobs/visdata/{job_id}")

    @staticmethod
    def get_manual_task_controller(client: "Itential", job_id: str, task_id: str) -> requests.Response:
        """
        Get a Manual Task's Controller (javascript function code)

        https://apidocs.itential.com/2020.2/api/app-workflow_engine/getManualTaskController/

        :param client: Itential client object. Passed in to all commands
        :param job_id: Returned when creating a job or querying for jobs by workflow name.
         Ex: "ec59ef85fef84e59bf36bd1e"
        :param task_id: ID of the task associated with the controller

        :return: requests.Response. Contains incoming and outgoing variables, error, and "TaskController"
         javascript code
        """
        return client.call(
            method="GET", url=f"{client.url}/workflow_engine/tasks/controller/job/{job_id}/task/{task_id}"
        )

    @staticmethod
    def get_task(client: "Itential", query: Dict[str, Any], filter: Dict[str, Any]) -> requests.Response:
        """
        Get the first Job's Task matching the query and return the data optionally modified by the filter.

        https://apidocs.itential.com/2020.2/api/-workflow_engine/getTask/

        :param client: Itential client object. Passed in to all commands
        :param query: An object of property names and values to use in the query against the tasks collection.
         The property name must exist in the tasks collection. There must be at least 1 field you query upon.
         Ex: {"someFieldName": "value expected in field"}
        :param filter: Filter specifying which fields are returned. The structure of each property must conform to
         'fieldName': 1. Ex: {"name": 1} will return the name field

        :return: requests.Response.
        """
        body = {"query": query, "filter": filter}
        return client.call(method="POST", url=f"{client.url}/workflow_engine/getTask", json=body)

    @staticmethod
    def get_task_details(client: "Itential", location: str, package: str, method: str) -> requests.Response:
        """
        Get the detailed information model for a task. Documentation for this method is incomplete.

        https://apidocs.itential.com/2020.2/api/-workflow_engine/getTaskDetails/

        :param client: Itential client object. Passed in to all commands
        :param location: Location: Application, Broker, Adapter
        :param package: Package name
        :param method: Method name

        :return: requests.Response.
         Ex: {
            "location": "Broker",
            "name": "query",
            "app": "WorkFlowEngine",
            "variables": {}
            }
        """
        return client.call(
            method="GET", url=f"{client.url}/workflow_engine/locations/{location}/packages/{package}/tasks/{method}"
        )

    @staticmethod
    def get_task_statuses(client: "Itential", job_id: str) -> requests.Response:
        """
        Get the status of each Task in a job by job ID.

        https://apidocs.itential.com/2020.2/api/-workflow_engine/getTaskStatuses/

        :param client: Itential client object. Passed in to all commands
        :param job_id: Returned when creating a job or querying for jobs by workflow name.
         Ex: "ec59ef85fef84e59bf36bd1e"

        :return: requests.Response Statuses of all Tasks in the job.
        """
        return client.call(method="GET", url=f"{client.url}/workflow_engine/job/statuses/{job_id}")

    @staticmethod
    def get_workflows_detailed_by_name(client: "Itential", workflow_name: str) -> requests.Response:
        """
        Get the status of each Task in a job by job ID.

        https://apidocs.itential.com/2020.2/api/-workflow_engine/getWorkflowsDetailedByName/

        :param client: Itential client object. Passed in to all commands
        :param workflow_name: Name of the workflow.
        :return: requests.Response Workflow details matching the Workflow name
        """
        return client.call(method="GET", url=f"{client.url}/workflow_engine/workflows/detailed/{workflow_name}")

    @staticmethod
    def is_active(client: "Itential") -> requests.Response:
        """
        Check if Staterator is currently active

        https://apidocs.itential.com/2020.2/api/-workflow_engine/isActive/

        :param client: Itential client object. Passed in to all commands
        :return: requests.Response Active flag for Staterator
        """
        return client.call(method="GET", url=f"{client.url}/workflow_engine/active")

    @staticmethod
    def list_job_groups(client: "Itential", job_id: str) -> requests.Response:
        """
        List the groups that have access to a job.

        https://apidocs.itential.com/2020.2/api/-workflow_engine/listJobGroups/

        :param client: Itential client object. Passed in to all commands
        :param job_id: Returned when creating a job or querying for jobs by workflow name.
         Ex: "ec59ef85fef84e59bf36bd1e"
        :return: requests.Response List of Group IDs.
        """
        return client.call(method="GET", url=f"{client.url}/workflow_engine/jobs/{job_id}/groups")

    @staticmethod
    def pause_job(client: "Itential", job_id: str) -> requests.Response:
        """
        Pause a job by job ID.

        https://apidocs.itential.com/2020.2/api/-workflow_engine/pauseJob/

        :param client: Itential client object. Passed in to all commands
        :param job_id: Returned when creating a job or querying for jobs by workflow name.
         Ex: "ec59ef85fef84e59bf36bd1e"
        :return: requests.Response Job after the pause.
        """
        body = {"job_id": job_id}
        return client.call(method="POST", url=f"{client.url}/workflow_engine/pauseJob", json=body)

    @staticmethod
    def prepare_metrics_logs(client: "Itential") -> requests.Response:
        """
        Compress the metrics logs directory for Jobs and return the link to download it.

        https://apidocs.itential.com/2020.2/api/-workflow_engine/prepareMetricsLogs/

        :param client: Itential client object. Passed in to all commands
        :return: requests.Response URL containing the compressed job metrics zipped file
        """
        return client.call(method="GET", url=f"{client.url}/workflow_engine/metrics/jobs")

    @staticmethod
    def query_jobs(client: "Itential", query: Dict[str, Any]) -> requests.Response:
        """
        Get jobs matching the query.

        https://apidocs.itential.com/2020.2/api/-workflow_engine/queryJobs/

        :param client: Itential client object. Passed in to all commands
        :param query: An object of property names and values to use in the query against the jobs collection.
         The property name must exist in the jobs collection. There must be at least 1 field you query upon.
         Ex: {"query": {"someFieldName": "Some Value to query the against the someFieldName property"}}
        :return: requests.Response All jobs matching the query.
        """
        body = {"query": query}
        return client.call(method="POST", url=f"{client.url}/workflow_engine/queryJobs", json=body)

    @staticmethod
    def query_tasks_brief(client: "Itential", query: Dict[str, Any]) -> requests.Response:
        """
        Get brief information about task(s) from the tasks collection in mongoDB.
        This will return only tasks that are attached to a currently running Job

        https://apidocs.itential.com/2020.2/api/-workflow_engine/queryTasksBrief/

        :param client: Itential client object. Passed in to all commands
        :param query: An object of property names and values to use in the query against the jobs collection.
         The property name must exist in the jobs collection. There must be at least 1 field you query upon.
         Ex: {"query": {"someFieldName": "Some Value to query the against the someFieldName property"}}
        :return: requests.Response Jobs' Tasks matching the query.
        """
        body = {"query": query}
        return client.call(method="POST", url=f"{client.url}/workflow_engine/queryTasksBrief", json=body)

    @staticmethod
    def release_task(client: "Itential", task_id: str) -> requests.Response:
        """
        Get brief information about task(s) from the tasks collection in mongoDB.
        This will return only tasks that are attached to a currently running Job

        https://apidocs.itential.com/2020.2/api/-workflow_engine/releaseTask/

        :param client: Itential client object. Passed in to all commands
        :param task_id: ID of a specific task. Ex: "cd34"
        :return: requests.Response Task details after releasing the Task.
        """
        body = {"task_id": task_id}
        return client.call(method="POST", url=f"{client.url}/workflow_engine/tasks/release", json=body)

    @staticmethod
    def remove_job_group(client: "Itential", job_id: str, group_id: str) -> requests.Response:
        """
        Remove a group from the list of authorized groups for a job.

        https://apidocs.itential.com/2020.2/api/-workflow_engine/removeJobGroup/

        :param client: Itential client object. Passed in to all commands
        :param job_id: Returned when creating a job or querying for jobs by workflow name.
         Ex: "ec59ef85fef84e59bf36bd1e"
        :param group_id: ID of a group. Ex:
        :return: requests.Response Status of the removal of a group from a job.
        """
        return client.call(method="DELETE", url=f"{client.url}/workflow_engine/jobs/{job_id}/groups/{group_id}")

    @staticmethod
    def replace_job_groups(client: "Itential", job_id: str, group_list: List[str]) -> requests.Response:
        """
        Overwrite the list of groups that have access to a job.

        https://apidocs.itential.com/2020.2/api/-workflow_engine/replaceJobGroups/

        :param client: Itential client object. Passed in to all commands
        :param job_id: Returned when creating a job or querying for jobs by workflow name.
         Ex: "ec59ef85fef84e59bf36bd1e"
        :param group_list: List of group IDs
        :return: requests.Response Status of the removal of a group from a job.
        """
        body = {"groups": group_list}
        return client.call(method="PUT", url=f"{client.url}/workflow_engine/jobs/{job_id}/groups", json=body)

    @staticmethod
    def resume_job(client: "Itential", job_id: str) -> requests.Response:
        """
        Resume a paused or errored job by job ID.

        https://apidocs.itential.com/2020.2/api/-workflow_engine/resumeJob/

        :param client: Itential client object. Passed in to all commands
        :param job_id: Returned when creating a job or querying for jobs by workflow name.
         Ex: "ec59ef85fef84e59bf36bd1e"
        :return: requests.Response Job after the resume.
        """
        body = {"job_id": job_id}
        return client.call(method="POST", url=f"{client.url}/workflow_engine/resumeJob", json=body)

    @staticmethod
    def revert_to_task(client: "Itential", job_id: str, current_task_id: str, target_task_id: str) -> requests.Response:
        """
        Revert a job from current Task to the target task by job ID and task names.

        https://apidocs.itential.com/2020.2/api/-workflow_engine/revertToTask/

        :param client: Itential client object. Passed in to all commands
        :param job_id: Returned when creating a job or querying for jobs by workflow name.
         Ex: "ec59ef85fef84e59bf36bd1e"
        :param current_task_id: Name of the current Task.
        :param target_task_id: Name of the Target Task.
        :return: requests.Response Job after the revert.
        """
        body = {"job_id": job_id, "current_task": current_task_id, "target_task": target_task_id}
        return client.call(method="POST", url=f"{client.url}/workflow_engine/revertToTask", json=body)

    @staticmethod
    def run_evaluation_group(
        client: "Itential", evaluation_group: List[Dict[str, Any]], all_true_flag: bool
    ) -> requests.Response:
        """
        Run a test evaluation.

        Note: This might be an accidentally exposed endpoint for the "evaluation" task within a workflow.

        https://apidocs.itential.com/2020.2/api/-workflow_engine/runEvaluationGroup/

        :param client: Itential client object. Passed in to all commands
        :param evaluation_group: Array of evaluations to run
        :param all_true_flag: All evaluation groups must pass, or not
        :return: requests.Response Result of evaluation.
        Ex:
            {
              "evaluation_group": [
                {
                  "operand_1": 47978204.94549683,
                  "operator": "contains",
                  "operand_2": 95772920.69505993,
                  "query": "somePropertyNameInOperand_1IfItIsAnObject"
                },
                {
                  "operand_1": [],
                  "operator": "contains",
                  "query": "somePropertyNameInOperand_1IfItIsAnObject"
                }
              ],
              "all_true_flag": true
            }
        """
        body = {
            "evaluation_group": evaluation_group,
            "all_true_flag": all_true_flag,
        }
        return client.call(method="POST", url=f"{client.url}/workflow_engine/runEvaluationGroup", json=body)

    @staticmethod
    def run_evaluation_groups(
        client: "Itential", evaluation_group: List[Dict[str, Any]], all_true_flag: bool
    ) -> requests.Response:
        """
        Run a test evaluation.

        Note: This might be an accidentally exposed endpoint for the "evaluation" task within a workflow.

        https://apidocs.itential.com/2020.2/api/-workflow_engine/runEvaluationGroups/

        :param client: Itential client object. Passed in to all commands
        :param evaluation_group: Array of evaluations to run
        :param all_true_flag: All evaluation groups must pass, or not
        :return: requests.Response Result of evaluation.
        Ex:
            {
              "evaluation_group": [
                {
                  "evaluations": [
                    {
                      "operand_1": true,
                      "operator": "contains",
                      "operand_2": "Lorem veniam",
                      "query": "somePropertyNameInOperand_1IfItIsAnObject"
                    },
                    {
                      "operand_1": [],
                      "operator": "contains",
                      "operand_2": -18752418,
                      "query": "somePropertyNameInOperand_1IfItIsAnObject"
                    }
                  ],
                  "all_true_flag": true
                }
              ],
              "all_true_flag": true
            }
        """
        body = {
            "evaluation_group": evaluation_group,
            "all_true_flag": all_true_flag,
        }
        return client.call(method="POST", url=f"{client.url}/workflow_engine/runEvaluationGroups", json=body)

    @staticmethod
    def run_validation(client: "Itential", workflow_json: Dict[str, Any]) -> requests.Response:
        """
        Validate a workflow, and return the resulting errors and warnings arrays.

        https://apidocs.itential.com/2020.2/api/-workflow_engine/runValidation/

        :param client: Itential client object. Passed in to all commands
        :param workflow_json: The workflow object to run validations against.
        :return: requests.Response Errors and warnings output from the validation process
        """
        body = {"workflow": workflow_json}
        return client.call(method="POST", url=f"{client.url}/workflow_engine/workflows/validate", json=body)

    @staticmethod
    def search_jobs(
        client: "Itential",
        expand: Optional[List[str]] = None,
        query: Optional[Dict[str, str]] = None,
        fields: Optional[Dict[str, int]] = None,
        local: Optional[bool] = True,
        sort: Optional[Dict[str, int]] = None,
        limit: int = 0,
        skip: int = 0,
    ) -> requests.Response:
        """
        TODO: Needs testing to identify what each param actually means. Documentation is very lacking.
        Search jobs with Options. This is similar to search_workflows, but with some additional fields.

        https://apidocs.itential.com/2020.2/api/app-workflow_engine/searchJobs/

        :param client: The Itential client state object
        :param expand: {"expand": ["user"]}.
        :param query: {"name": "workflow_name"}. "name" field can be any field in the workflow json.
        :param fields: {"name": 1}. "name" field can be any field in the workflow json.
         Used to filter out other fields
        :param sort: {"name": -1} or {"name": 1}. "name" field can be any field in the workflow json.
        :param limit: Max results to return
        :param local:
        :param skip: Number to offset the search by
        :return: requests.Response example json(): Search Results: {"results": ["_id": int, "name": str], "skip": int,
         "limit": int, "total": int}
        Ex: {
              "options": {
                "expand": [
                  "user",
                  "owner",
                  "owner",
                  "user"
                ],
                "fields": {
                  "name": 1
                },
                "query": {
                  "name": "abcd"
                },
                "limit": 50,
                "local": true,
                "skip": 0,
                "sort": {
                  "name": -1
                }
              }
            }
        """

        data = {"options": {"fields": {"name": 1}, "limit": limit, "local": local, "skip": skip, "sort": {"name": 1}}}

        if query:
            data["options"]["query"] = query

        if expand:
            data["options"]["expand"] = expand

        if fields:
            data["options"]["fields"] = fields

        if sort:
            data["options"]["sort"] = sort

        return client.call(method="POST", url=rf"{client.url}/workflow_engine/jobs/search", json=data)

    @staticmethod
    def search_tasks(
        client: "Itential",
        filter: Dict[str, Any],
        expand: Optional[List[str]] = None,
        query: Optional[Dict[str, str]] = None,
        fields: Optional[Dict[str, int]] = None,
        local: Optional[bool] = True,
        sort: Optional[Dict[str, int]] = None,
        limit: int = 0,
        skip: int = 0,
    ) -> requests.Response:
        """
        TODO: Needs testing to identify what each param actually means. Documentation is very lacking.
        Search jobs with Options.

        Note: Unique in that it has an additional parameter "filter".

        https://apidocs.itential.com/2020.2/api/app-workflow_engine/searchTasks/

        :param client: The Itential client state object
        :param filter: Search Filter Ex: {"someFieldName": "Some Value to query the against the someFieldName property"}
        :param expand: {"expand": ["user"]}.
        :param query: {"name": "workflow_name"}. "name" field can be any field in the workflow json.
        :param fields: {"name": 1}. "name" field can be any field in the workflow json.
         Used to filter out other fields
        :param sort: {"name": -1} or {"name": 1}. "name" field can be any field in the workflow json.
        :param limit: Max results to return
        :param local:
        :param skip: Number to offset the search by
        :return: requests.Response example json(): Search Results: {"results": ["_id": int, "name": str], "skip": int,
         "limit": int, "total": int}
        Ex: {
              "filter": {
                "someFieldName": "Some Value to query the against the someFieldName property"
              },
              "options": {
                "expand": [
                  "owner",
                  "owner",
                  "created_by",
                  "owner",
                  "owner"
                ],
                "fields": {
                  "name": 1
                },
                "query": {
                  "name": "abcd"
                },
                "limit": 50,
                "local": false,
                "skip": 0,
                "sort": {
                  "name": -1
                }
              }
            }
        """

        data = {
            "filter": filter,
            "options": {"fields": {"name": 1}, "limit": limit, "local": local, "skip": skip, "sort": {"name": 1}},
        }

        if query:
            data["options"]["query"] = query

        if expand:
            data["options"]["expand"] = expand

        if fields:
            data["options"]["fields"] = fields

        if sort:
            data["options"]["sort"] = sort

        return client.call(method="POST", url=rf"{client.url}/workflow_engine/tasks/search", json=data)

    @staticmethod
    def search_workflows(
        client: "Itential",
        query: Optional[Dict[str, str]] = None,
        fields: Optional[Dict[str, int]] = None,
        sort: Optional[Dict[str, int]] = None,
        limit: int = 0,
        skip: int = 0,
    ) -> requests.Response:
        """
        Search Workflows with Options

        https://apidocs.itential.com/2020.2/api/app-workflow_engine/searchWorkflows/

        :param client: The Itential client state object
        :param query: {"name": "workflow_name"}. "name" field can be any field in the workflow json.
        :param fields: {"name": 1}. "name" field can be any field in the workflow json.
         Used to filter out other fields
        :param sort: {"name": -1} or {"name": 1}. "name" field can be any field in the workflow json.
        :param limit: Max results to return
        :param skip: Number to offset the search by
        :return: requests.Response example json(): {"results": ["_id": int, "name": str], "skip": int,
         "limit": int, "total": int}
        """

        data = {"options": {"fields": {"name": 1}, "limit": limit, "skip": skip, "sort": {"name": 1}}}

        if query:
            data["options"]['query'] = query

        if fields:
            data["options"]["fields"] = fields

        if sort:
            data["options"]["sort"] = sort

        return client.call(method="POST", url=rf"{client.url}/workflow_engine/workflows/search", json=data)

    @staticmethod
    def unwatch_job(client: "Itential", job_id: str) -> requests.Response:
        """
        Remove the current user from the watchers list of a job.

        https://apidocs.itential.com/2020.2/api/-workflow_engine/unwatchJob/

        :param client: Itential client object. Passed in to all commands
        :param job_id: ID of the job
        :return: requests.Response User id removed from the watchers list.
        """
        return client.call(method="DELETE", url=f"{client.url}/workflow_engine/job/{job_id}/watch")

    @staticmethod
    def validate_all_loops(client: "Itential", workflow_json: Dict[str, Any]) -> requests.Response:
        """
        Validate all looped Tasks in a Workflow by Workflow details.

        https://apidocs.itential.com/2020.2/api/-workflow_engine/validateAllLoops/

        :param client: Itential client object. Passed in to all commands
        :param workflow_json: Full workflow json/dict
        :return: requests.Response User id removed from the watchers list.
        """
        body = {"workflow_details": workflow_json}
        return client.call(method="POST", url=f"{client.url}/workflow_engine/validateAllLoops", json=body)

    @staticmethod
    def watch_job(client: "Itential", job_id: str) -> requests.Response:
        """
        Add current user to a job's watchers list.

        https://apidocs.itential.com/2020.2/api/-workflow_engine/watchJob/

        :param client: Itential client object. Passed in to all commands
        :param job_id: ID of the job
        :return: requests.Response User id added to the watchers list.
        """
        return client.call(method="PUT", url=f"{client.url}/workflow_engine/job/{job_id}/watch")



"""
Link to Itential Docs: https://apidocs.itential.com/2020.2/api/app-workflow_engine/

Implemented  Doc String  Tests
    [ ]         [ ]       [ ]   createAccount
    [ ]         [ ]       [ ]   createAdapter
    [ ]         [ ]       [ ]   createGroup
    [ ]         [ ]       [ ]   createIntegration
    [ ]         [ ]       [ ]   createIntegrationModel
    [ ]         [ ]       [ ]   createProfile
    [ ]         [ ]       [ ]   deleteAccount
    [ ]         [ ]       [ ]   deleteAdapter
    [ ]         [ ]       [ ]   deleteGroup
    [ ]         [ ]       [ ]   deleteIntegration
    [ ]         [ ]       [ ]   deleteIntegrationModel
    [ ]         [ ]       [ ]   deleteProfile
    [ ]         [ ]       [ ]   exportPrebuilt
    [ ]         [ ]       [ ]   getAccount
    [ ]         [ ]       [ ]   getAccounts
    [ ]         [ ]       [ ]   getAdapter
    [ ]         [ ]       [ ]   getAdapterChangelogs
    [ ]         [ ]       [ ]   getAdapterHealth
    [ ]         [ ]       [ ]   getAdapterModelTypes
    [ ]         [ ]       [ ]   getAdapters
    [ ]         [ ]       [ ]   getAdapterSchema
    [ ]         [ ]       [ ]   getAdaptersHealth
    [ ]         [ ]       [ ]   getApplication
    [ ]         [ ]       [ ]   getApplicationChangelogs
    [ ]         [ ]       [ ]   getApplicationHealth
    [ ]         [ ]       [ ]   getApplications
    [ ]         [ ]       [ ]   getApplicationSchema
    [ ]         [ ]       [ ]   getApplicationsHealth
    [ ]         [ ]       [ ]   getAssignableRoles
    [ ]         [ ]       [ ]   getGroup
    [ ]         [ ]       [ ]   getGroups
    [ ]         [ ]       [ ]   getIntegration
    [ ]         [ ]       [ ]   getIntegrationModel
    [ ]         [ ]       [ ]   getIntegrationModels
    [ ]         [ ]       [ ]   getIntegrations
    [ ]         [ ]       [ ]   getIntegrationSchema
    [ ]         [ ]       [ ]   getPrebuilt
    [ ]         [ ]       [ ]   getPrebuilts
    [ ]         [ ]       [ ]   getProfile
    [ ]         [ ]       [ ]   getProfiles
    [ ]         [ ]       [ ]   getProfileSchema
    [ ]         [ ]       [ ]   getRepositoryPrebuilts
    [ ]         [ ]       [ ]   getServerHealth
    [ ]         [ ]       [ ]   getSystemHealth
    [ ]         [ ]       [ ]   importPrebuilt
    [ ]         [ ]       [ ]   removePrebuilt
    [ ]         [ ]       [ ]   restartAdapter
    [ ]         [ ]       [ ]   restartApplication
    [ ]         [ ]       [ ]   setAccountGroups
    [ ]         [ ]       [ ]   setAccountPassword
    [ ]         [ ]       [ ]   setGroupRoles
    [ ]         [ ]       [ ]   startAdapter
    [ ]         [ ]       [ ]   startApplication
    [ ]         [ ]       [ ]   stopAdapter
    [ ]         [ ]       [ ]   stopApplication
    [ ]         [ ]       [ ]   switchActiveProfile
    [ ]         [ ]       [ ]   updateAdapter
    [ ]         [ ]       [ ]   updateAdapterLogging
    [ ]         [ ]       [ ]   updateAdapterProperties
    [ ]         [ ]       [ ]   updateApplication
    [ ]         [ ]       [ ]   updateApplicationLogging
    [ ]         [ ]       [ ]   updateApplicationProperties
    [ ]         [ ]       [ ]   updateIntegration
    [ ]         [ ]       [ ]   updateIntegrationModel
    [ ]         [ ]       [ ]   updateIntegrationProperties
    [ ]         [ ]       [ ]   updatePrebuilt
    [ ]         [ ]       [ ]   updateProfile
    [ ]         [ ]       [ ]   validateIntegrationModel
    [ ]         [ ]       [ ]   validatePrebuilt
"""



"""
Link to Itential Docs: https://apidocs.itential.com/2020.2/api/app-task_worker/

Implemented  Doc String  Tests
    [ ]         [ ]       [ ]   activate
    [ ]         [ ]       [ ]   deactivate
    [ ]         [ ]       [ ]   isActive
"""



from itential.modules.v2020_2._app_form_builder import *
from itential.modules.v2020_2._adapter_automation_gateway import *
from itential.modules.v2020_2._app_admin_essentials import *
from itential.modules.v2020_2._app_ag_manager import *
from itential.modules.v2020_2._app_automation_catalog import *
from itential.modules.v2020_2._app_automation_studio import *
from itential.modules.v2020_2._app_configuration_manager import *
from itential.modules.v2020_2._app_form_builder import *
from itential.modules.v2020_2._app_json_forms import *
from itential.modules.v2020_2._app_jst import *
from itential.modules.v2020_2._app_mop import *
from itential.modules.v2020_2._app_nso_manager import *
from itential.modules.v2020_2._app_operations_manager import *
from itential.modules.v2020_2._app_service_catalog import *
from itential.modules.v2020_2._app_service_management import *
from itential.modules.v2020_2._app_template_builder import *
from itential.modules.v2020_2._app_workflow_builder import *
from itential.modules.v2020_2._app_workflow_engine import *
from itential.modules.v2020_2._pronghorn_core import *



"""
Link to Itential Docs: https://apidocs.itential.com/2020.2/api/app-mop/

Implemented  Doc String  Tests
    [ ]         [ ]       [ ]   chooseDevice
    [ ]         [ ]       [ ]   createAnalyticTemplate
    [ ]         [ ]       [ ]   createTemplate
    [ ]         [ ]       [ ]   deleteAnalyticTemplate
    [ ]         [ ]       [ ]   deleteTemplate
    [ ]         [ ]       [ ]   diffConfig
    [ ]         [ ]       [ ]   exportTemplate
    [ ]         [ ]       [ ]   GetBootFlash
    [ ]         [ ]       [ ]   getDeviceObjectsFiltered
    [ ]         [ ]       [ ]   getDevicesFiltered
    [ ]         [ ]       [ ]   getDevicesFilteredDetailedResults
    [ ]         [ ]       [ ]   getDiff
    [ ]         [ ]       [ ]   importTemplate
    [ ]         [ ]       [ ]   listAnalyticTemplates
    [ ]         [ ]       [ ]   listAnAnalyticTemplate
    [ ]         [ ]       [ ]   listATemplate
    [ ]         [ ]       [ ]   listTemplates
    [ ]         [ ]       [ ]   passThru
    [ ]         [ ]       [ ]   reattempt
    [ ]         [ ]       [ ]   reloadFailed
    [ ]         [ ]       [ ]   reviewSummary
    [ ]         [ ]       [ ]   runAnalyticsTemplate
    [ ]         [ ]       [ ]   runAnalyticsTemplateDevices
    [ ]         [ ]       [ ]   RunCommand
    [ ]         [ ]       [ ]   RunCommandDevices
    [ ]         [ ]       [ ]   RunCommandTemplate
    [ ]         [ ]       [ ]   RunCommandTemplateSingleCommand
    [ ]         [ ]       [ ]   runTemplatesDiff
    [ ]         [ ]       [ ]   updateAnalyticTemplate
    [ ]         [ ]       [ ]   updateTemplate
    [ ]         [ ]       [ ]   verifyConfig
    [ ]         [ ]       [ ]   viewTemplateResults
"""



"""
Link to Itential Docs: https://apidocs.itential.com/2020.2/api/app-automation_studio/

Implemented  Doc String  Tests
    [x]         [x]       [ ]   createAutomation
    [x]         [x]       [ ]   createTemplate
    [x]         [x]       [ ]   deleteTemplate
    [x]         [x]       [ ]   getMethodOptions
    [x]         [x]       [ ]   getTemplate
    [x]         [x]       [ ]   getTemplates
    [x]         [x]       [ ]   importAutomations
    [x]         [x]       [ ]   importTemplates
    [x]         [x]       [ ]   runTransformation
    [x]         [x]       [ ]   updateAutomation
    [x]         [x]       [ ]   updateTemplate
"""
from typing import TYPE_CHECKING, List, Dict, Any

if TYPE_CHECKING:
    from itential.core import Itential
    import requests


class AppAutomationStudio:
    """https://apidocs.itential.com/2020.2/api/app-automation_studio/createAutomation/"""

    @staticmethod
    def create_automation(client: "Itential", automation_json: Dict[str, Any]) -> "requests.Response":
        """
        Creates a new automation document.
        https://apidocs.itential.com/2020.2/api/app-automation_studio/createAutomation/
        :param client: Itential state object
        :param automation_json: json representation of the Automation to create
        :return: requests.Response: Created automation and associated edit URI
        """
        body = {"automation": automation_json}
        return client.call(method="POST", url=f"{client.url}/automation-studio/automations", json=body)

    @staticmethod
    def create_template(client: "Itential", template_json: Dict[str, Any]) -> "requests.Response":
        """
        Creates a new template document.
        https://apidocs.itential.com/2020.2/api/app-automation_studio/createTemplate/
        :param client: Itential state object
        :param template_json: json representation of the Template to create
        :return: requests.Response: Created template and associated edit URI.
        """
        body = {"template": template_json}
        return client.call(method="POST", url=f"{client.url}/automation-studio/templates", json=body)

    @staticmethod
    def delete_template(client: "Itential", template_id: str) -> "requests.Response":
        """
        Deletes a template document.
        https://apidocs.itential.com/2020.2/api/app-automation_studio/deleteTemplate/
        :param client: Itential state object
        :param template_id: Unique template ID
        :return: requests.Response:
        """
        return client.call(method="DELETE", url=f"{client.url}/automation-studio/templates/{template_id}")

    @staticmethod
    def get_method_options(client: "Itential") -> "requests.Response":
        """
        Get all available rest calls in IAP.
        https://apidocs.itential.com/2020.2/api/app-automation_studio/getMethodOptions/
        :param client: Itential state object
        :return: requests.Response: List of all rest calls available in the system.
        """
        return client.call(method="GET", url=f"{client.url}/automation-studio/json-forms/method-options")

    @staticmethod
    def get_template(client: "Itential", template_id: str, query_parameters: Dict[str, Any]) -> "requests.Response":
        """
        Gets a single template document.
        https://apidocs.itential.com/2020.2/api/app-automation_studio/getTemplate/
        :param client: Itential state object
        :param template_id: ObjectId specifying a template entity.
        :param query_parameters: Optional parameters for projecting fields in the template document.
            (example in DOM only)
        :return: requests.Response:
        """
        params = {"queryParameters": query_parameters}
        return client.call(method="GET", url=f"{client.url}/automation-studio/templates/{template_id}", params=params)

    @staticmethod
    def get_templates(client: "Itential", query_parameters: Dict[str, Any]) -> "requests.Response":
        """
        Gets a page of template documents.
        https://apidocs.itential.com/2020.2/api/app-automation_studio/getTemplates/
        :param client: Itential state object
        :param query_parameters: Parameters for filtering, paginating, projecting, and sorting template documents.
            (example in DOM only)
        :return: requests.Response:
        """
        params = {"queryParameters": query_parameters}
        return client.call(method="GET", url=f"{client.url}/automation-studio/templates", params=params)

    @staticmethod
    def import_automations(client: "Itential", automation_list: List[Dict[str, Any]]) -> "requests.Response":
        """
        Imports a new automation document.
        https://apidocs.itential.com/2020.2/api/app-automation_studio/importAutomations/
        :param client: Itential state object
        :param automation_list: List of automation json representations.
        :return: requests.Response: Results from each individual import operation.
        """
        body = {"automations": automation_list}
        return client.call(method="POST", url=f"{client.url}/automation-studio/automations/import", json=body)

    @staticmethod
    def import_templates(client: "Itential", template_list: List[Dict[str, Any]]) -> "requests.Response":
        """
        Imports a new template document.
        https://apidocs.itential.com/2020.2/api/app-automation_studio/importTemplates/
        :param client: Itential state object
        :param template_list: List of template json representations.
        :return: requests.Response: Results from each individual import operation.
        """
        body = {"templates": template_list}
        return client.call(method="POST", url=f"{client.url}/automation-studio/templates/import", json=body)

    @staticmethod
    def run_transformation(
        client: "Itential", transformation_id: str, transformation_data: Dict[str, Any]
    ) -> "requests.Response":
        """
        Transforms data using JST transformation document.
        https://apidocs.itential.com/2020.2/api/app-automation_studio/runTransformation/
        :param client: Itential state object
        :param transformation_data: Transformation json representation
        :param transformation_id: Unique transformation ID
        :return: requests.Response: Transformed data.
        """
        body = {"data": transformation_data}
        return client.call(
            method="POST",
            url=f"{client.url}/automation-studio/json-forms/runTransformation/{transformation_id}",
            json=body,
        )

    @staticmethod
    def update_automation(
        client: "Itential", automation_id: str, automation_update: Dict[str, Any]
    ) -> "requests.Response":
        """
        Replaces a automation document.
        https://apidocs.itential.com/2020.2/api/app-automation_studio/updateAutomation/
        :param client: Itential state object
        :param automation_id: Unique automation ID
        :param automation_update: Complete automation definition/json representation to replace the
            existing automation document with.
        :return: requests.Response:
        """
        body = {"update": automation_update}
        return client.call(method="PUT", url=f"{client.url}/automation-studio/automations/{automation_id}", json=body)

    @staticmethod
    def update_template(client: "Itential", template_id: str, template_update: Dict[str, Any]) -> "requests.Response":
        """
        Replaces a template document.
        https://apidocs.itential.com/2020.2/api/app-automation_studio/updateTemplate/
        :param client: Itential state object
        :param template_id: Unique template ID
        :param template_update: Complete template definition to replace the existing template document with.
            May not contain fields '_id', 'created', 'createdBy', 'lastUpdated', or 'lastUpdatedBy'.
        :return: requests.Response:
        """
        body = {"update": template_update}
        return client.call(method="PUT", url=f"{client.url}/automation-studio/templates/{template_id}", json=body)



# from itential.modules.version2020._app_form_builder import AppFormBuilder2020
#
#
# class AppFormBuilder2021(AppFormBuilder2020):
#     def build_form_meth(*args, **kwargs):
#         print('DEPRECIATED build form meth 2021. '
#               'New method is "cool_build_form_func')
#
#     def cool_build_form_meth(*args, **kwargs):
#         print('cool build form func 2021')



"""
Link to Itential Docs: https://apidocs.itential.com/2020.2/api/app-json_forms/

Implemented  Doc String  Tests
    [ ]         [ ]       [ ]   createForm
    [ ]         [ ]       [ ]   decodeSchema
    [ ]         [ ]       [ ]   deleteForms
    [ ]         [ ]       [ ]   getFormById
    [ ]         [ ]       [ ]   getForms
    [ ]         [ ]       [ ]   importForms
    [ ]         [ ]       [ ]   RenderJsonSchema
    [ ]         [ ]       [ ]   RenderYangForm
    [ ]         [ ]       [ ]   ShowJsonForm
    [ ]         [ ]       [ ]   updateForm
    [ ]         [ ]       [ ]   validateData
    [ ]         [ ]       [ ]   validateForm
    [ ]         [ ]       [ ]   yangToSchema
"""



"""
Link to Itential Docs: https://apidocs.itential.com/2020.2/api/app-service_management/

Implemented  Doc String  Tests
    [ ]         [ ]       [ ]   checkSync
    [ ]         [ ]       [ ]   config
    [ ]         [ ]       [ ]   createServiceModelForm
    [ ]         [ ]       [ ]   deleteInstances
    [ ]         [ ]       [ ]   deleteServicePathsDryRun
    [ ]         [ ]       [ ]   deviceModifications
    [ ]         [ ]       [ ]   getDevicesInServiceInstance
    [ ]         [ ]       [ ]   getInstance
    [ ]         [ ]       [ ]   getInstancesOfService
    [ ]         [ ]       [ ]   getServiceInstanceMap
    [ ]         [ ]       [ ]   getServiceModel
    [ ]         [ ]       [ ]   getServiceModelMap
    [ ]         [ ]       [ ]   getServicesDetails
    [ ]         [ ]       [ ]   listServiceModels
    [ ]         [ ]       [ ]   reactiveRedeploy
    [ ]         [ ]       [ ]   saveInstances
    [ ]         [ ]       [ ]   testInstances
    [ ]         [ ]       [ ]   ViewServiceModelForm
    [ ]         [ ]       [ ]   ViewTestService
"""



"""
Link to Itential Docs: https://apidocs.itential.com/2020.2/api/app-service_catalog/

Implemented  Doc String  Tests
    [ ]         [ ]       [ ]   AddNewServiceToCatalog
    [ ]         [ ]       [ ]   deleteService
    [ ]         [ ]       [ ]   GetFormData
    [ ]         [ ]       [ ]   GetFormId
    [ ]         [ ]       [ ]   GetForms
    [ ]         [ ]       [ ]   getGroups
    [ ]         [ ]       [ ]   GetUserObject
    [ ]         [ ]       [ ]   GetWorkflows
    [ ]         [ ]       [ ]   invokeServiceOrder
    [ ]         [ ]       [ ]   ServiceCatalogStore
    [ ]         [ ]       [ ]   ServiceModels
    [ ]         [ ]       [ ]   UpdateServiceInCatalog
"""



"""
Link to Itential Docs: https://apidocs.itential.com/2020.2/api/app-form_builder/

Implemented  Doc String  Tests
    [ ]         [ ]       [ ]   createFormGroupEntry
    [ ]         [ ]       [ ]   deleteForm
    [ ]         [ ]       [ ]   deleteFormGroups
    [ ]         [ ]       [ ]   exportForm
    [ ]         [ ]       [ ]   fetchData
    [ ]         [ ]       [ ]   getElementDefinition
    [ ]         [ ]       [ ]   getForm
    [ ]         [ ]       [ ]   getFormByName
    [ ]         [ ]       [ ]   importForm
    [ ]         [ ]       [ ]   listElements
    [ ]         [ ]       [ ]   listFormGroups
    [ ]         [ ]       [ ]   listForms
    [ ]         [ ]       [ ]   preserveFormData
    [ ]         [ ]       [ ]   ProviderForm
    [ ]         [ ]       [ ]   removeFormGroup
    [ ]         [ ]       [ ]   replaceFormGroups
    [ ]         [ ]       [ ]   saveForm
    [ ]         [ ]       [ ]   searchForms
    [ ]         [ ]       [ ]   ShowFormByName
    [ ]         [ ]       [ ]   ShowFormPreserveData
"""



"""
This app contains all of the methods used to modify transformations/JSTs on the server.
Link to Itential Docs: https://apidocs.itential.com/2020.2/api/app-jst/

Implemented  Doc String  Tests
    [x]         [x]       [x]   createTransformation
    [x]         [x]       [x]   deleteTransformation
    [x]         [x]       [ ]   getTransformation
    [x]         [x]       [ ]   importTransformation
    [x]         [x]       [ ]   runTransformation
    [x]         [x]       [ ]   searchTransformations
    [x]         [x]       [ ]   updateTransformation

"""
import json
from typing import TYPE_CHECKING, Dict, Any, Optional

if TYPE_CHECKING:
    from itential.core import Itential
    import requests


class AppJst:
    """https://apidocs.itential.com/2020.2/api/app-jst"""

    @staticmethod
    def create_transformation(client: "Itential", jst_obj: Dict[str, Any]) -> "requests.Response":
        """
        Creates a single transformation.
        https://apidocs.itential.com/2020.2/api/app-jst/createTransformation/ (Misleading param example and schema)
        :param client: Itential state object
        :param jst_obj: Json object representation of the JST
        :return: requests.Response
        """
        return client.call(method="POST", url=f"{client.url}/transformations", json=jst_obj)

    @staticmethod
    def delete_transformation(client: "Itential", jst_id: str) -> "requests.Response":
        """
        Deletes a single transformation from the server.
        https://apidocs.itential.com/2020.2/api/app-jst/deleteTransformation/
        :param client: Itential state object
        :param jst_id: "_id" field of the JST json. ex: 6230fd0b56192935eed3eeb8
        :return: requests.Response  200: empty string
        """
        return client.call(method="DELETE", url=f"{client.url}/transformations/{jst_id}")

    @staticmethod
    def get_transformations(client: "Itential", jst_id: str) -> "requests.Response":
        """
        Returns all transformations from the server, in full. This can return a rather large payload.
        Documentation is incorrect. Doesn't require "queryParameters". Don't have any examples of queryParameters usage.
        https://apidocs.itential.com/2021.2/api/app-jst/searchTransformations/
        :param client: Itential state object
        :param jst_id: "_id" field of the JST json. ex: 6230fd0b56192935eed3eeb8
        :return: requests.Response
        """
        return client.call(method="GET", url=rf"{client.url}/transformations/{jst_id}")

    @staticmethod
    def import_transformation(client: "Itential", jst_obj: Dict[str, Any]) -> "requests.Response":
        """
        Imports a single transformation/JST to the server. Will duplicate files if JST already exists.
        Will append a " (n)" to the end of the JST name. "n" is an incrementing integer
        https://apidocs.itential.com/2020.2/api/app-jst/importTransformation/ (Misleading param example and schema)
        :param client: Itential state object
        :param jst_obj: The json object representation of the transformation/JST
        :return: requests.Response
        """
        return client.call(method="POST", url=f"{client.url}/transformations/import", json=jst_obj)

    @staticmethod
    def run_transformation(
        client: "Itential", jst_id: str, incoming: Dict[str, Any], options: Optional[Dict[str, Any]] = None
    ) -> "requests.Response":
        """
        Runs a transformation
        https://apidocs.itential.com/2020.2/api/app-jst/runTransformation/
        :param client: Itential state object
        :param jst_id: "_id" of the JST. ex: 6143988e49bd502787711378
        :param incoming: A dict object of the input schema.
        :param options: < Unclear > This is not required to run a transformation.
        :return: requests.Response Returns the outgoing schema as a json obj
        """
        data = {"incoming": incoming, "options": options if options else {}}
        return client.call(method="POST", url=f"{client.url}/transformations/{jst_id}", json=data)

    @staticmethod
    def search_transformations(client: "Itential") -> "requests.Response":
        """
        Returns all transformations from the server, in full. This can return a rather large payload.
        Doesn't require "queryParameters". The usage found in the website DOM is not correct
        Documentation is incomplete. Requires further testing to figure this one out.
        https://apidocs.itential.com/2021.2/api/app-jst/searchTransformations/
        :param client: Itential state object
        :param query_parameters: < Unclear > Omitted the implementation at this time.
        :return: requests.Response
        """
        return client.call(method="GET", url=rf"{client.url}/transformations")

    @staticmethod
    def update_transformation(client: "Itential", jst_id: str, jst_obj: Dict[str, Any]) -> "requests.Response":
        """
        Updates/Overwrites an existing transformation/JST based on ID. Does not create a duplicate file.
        https://apidocs.itential.com/2020.2/api/app-jst/updateTransformation/  (Misleading param example and schema)
        :param client: Itential state object
        :param jst_id: "_id" field of the JST json. ex: 6230fd0b56192935eed3eeb8
        :param jst_obj: Full JST json. Does not require {"transformation": JST_json} like docs show.
        :return: requests.Response
        """
        return client.call(method="PUT", url=f"{client.url}/transformations/{jst_id}", json=jst_obj)
